import curses
import time
from enum import Enum
from typing import List, Callable

import py_cui

from qrogue.game.logic.actors import Boss, Controllable, Enemy, Riddle, Robot
from qrogue.game.world.map import CallbackPack, LevelMap, SpaceshipMap, WorldMap
from qrogue.game.world.navigation import Direction
from qrogue.game.world.tiles import WalkTriggerTile, Message, Collectible
from qrogue.graphics.rendering import MultiColorRenderer
from qrogue.graphics.popups import Popup, MultilinePopup, ConfirmationPopup
from qrogue.graphics.widgets import Renderable, SpaceshipWidgetSet, BossFightWidgetSet, ExploreWidgetSet, \
    FightWidgetSet, MenuWidgetSet, MyWidgetSet, NavigationWidgetSet, PauseMenuWidgetSet, RiddleWidgetSet, \
    ShopWidgetSet, WorkbenchWidgetSet
from qrogue.util import achievements, common_messages, CheatConfig, ColorConfig, Config, GameplayConfig, HelpText, \
    HelpTextType, Logger, PathConfig, MapConfig, Controls, Keys
from qrogue.util.game_simulator import GameSimulator
from qrogue.util.key_logger import KeyLogger

from qrogue.management import MapManager, Pausing, SaveData


class QrogueCUI(py_cui.PyCUI):
    def __init__(self, seed: int, width: int = 8, height: int = 9):
        super().__init__(width, height)
        self.set_title(f"Qrogue {Config.version()}")
        self.__seed = seed

        # init management
        Logger.instance().set_popup(self.show_message_popup, self.show_error_popup)
        CheatConfig.init(self.__show_message_popup, self.__show_input_popup)
        Popup.update_popup_functions(self.__show_message_popup)
        ConfirmationPopup.update_popup_function(self.__show_confirmation_popup)

        Pausing(self.__pause_game)
        CallbackPack(self.__start_level, self.__start_fight, self.__start_boss_fight,
                           self.__open_riddle, self.__visit_shop, self.__game_over)
        SaveData()  # todo load data
        MapManager(seed, self.__show_world, self.__start_level)
        Popup.update_check_achievement_function(SaveData.instance().achievement_manager.check_achievement)
        common_messages.set_show_callback(Popup.message)
        common_messages.set_ask_callback(ConfirmationPopup.ask)
        WalkTriggerTile.set_show_explanation_callback(Popup.from_message)
        Message.set_show_callback(Popup.from_message)
        Collectible.set_pickup_message_callback(Popup.message)

        self.__key_logger = None  #KeyLogger(seed)
        self.__simulator = None
        self.__state_machine = StateMachine(self)
        self.__controls = Controls()   # todo load later from file!
        self.__last_input = time.time()
        self.__last_key = None
        self.__focused_widget = None

        # init widget sets
        self.__menu = MenuWidgetSet(self.__controls, self.__render, Logger.instance(), self, self.__start_playing,
                                    self.stop, self.__choose_simulation)
        self.__pause = PauseMenuWidgetSet(self.__controls, self.__render, Logger.instance(), self,
                                          self.__general_continue, self.switch_to_menu)

        self.__spaceship = SpaceshipWidgetSet(self.__controls, Logger.instance(), self, self.__render)
        self.__workbench = WorkbenchWidgetSet(self.__controls, Logger.instance(), self,
                                              SaveData.instance().available_robots(), self.__render,
                                              self.__continue_spaceship)
        self.__navigation = NavigationWidgetSet(self.__controls, Logger.instance(), self, self.__render)

        self.__explore = ExploreWidgetSet(self.__controls, self.__render, Logger.instance(), self)
        self.__fight = FightWidgetSet(self.__controls, self.__render, Logger.instance(), self, self.__continue_explore,
                                      self.__game_over)
        self.__boss_fight = BossFightWidgetSet(self.__controls, self.__render, Logger.instance(), self,
                                               self.__continue_explore, self.__game_over, self.__won_tutorial)
        self.__riddle = RiddleWidgetSet(self.__controls, self.__render, Logger.instance(), self, self.__continue_explore)
        self.__shop = ShopWidgetSet(self.__controls, self.__render, Logger.instance(), self, self.__continue_explore)

        self.__cur_widget_set = None
        self.__init_keys()

        self.__state_machine.change_state(State.Menu, None)
        self.__game_started = False

        def open_world_view(direction: Direction, controllable: Controllable):
            return MapManager.instance().load_map(MapConfig.hub_world(), None)

        def start_test_level(direction: Direction, controllable: Controllable):
            return MapManager.instance().load_map(MapConfig.test_level(), None)

        self.__spaceship_map = SpaceshipMap(seed, SaveData.instance().player, open_world_view, self.__use_workbench,
                                            start_test_level)

    def _refresh_height_width(self) -> None:
        try:
            super(QrogueCUI, self)._refresh_height_width()
        except py_cui.errors.PyCUIOutOfBoundsError:
            print("[Qrogue] ERROR!")
            rows, cols = self._grid.get_dimensions_absolute()
            x, y = self._grid.get_dimensions()
            x, y = (x * 3 + 1, y * 3 + 1)
            print(f"Current dimensions are ({cols}, {rows}) but at least ({x}, {y}) is needed. "
                  f"We recommend to make it more wide than high though, e.g. ({x * 3}, {y}) would be a suitable size. "
                  f"Alternatively you can also reduce the font size.")
            print("If you are running it in Windows Powershell you can also try to press ALT+ENTER to change to "
                  "fullscreen mode.")
            raise py_cui.errors.PyCUIOutOfBoundsError

    @property
    def is_simulating(self) -> bool:
        return self.__simulator is not None

    def start(self):
        self.render()
        super(QrogueCUI, self).start()

    def __choose_simulation(self):
        title = "Enter the path to the .qrkl-file to simulate:"
        self.__show_input_popup(title, py_cui.WHITE_ON_CYAN, self.__start_simulation)

    def __start_simulation(self, path: str):
        if not path.endswith(".qrkl"):
            path += ".qrkl"
        try:
            self.__simulator = GameSimulator(self.__controls, path, in_keylog_folder=True)
            self.__menu.simulate_with_seed(self.__simulator.seed)
            # go back to the original position of the cursor and start the game
            super(QrogueCUI, self)._handle_key_presses(self.__controls.get_key(Keys.SelectionUp))
            super(QrogueCUI, self)._handle_key_presses(self.__controls.get_key(Keys.SelectionUp))
            super(QrogueCUI, self)._handle_key_presses(self.__controls.get_key(Keys.Action))
            if self.__simulator.version == Config.version():
                __space = "Space"
                Popup.message("Starting Simulation", f"You started a run with \nseed = {self.__simulator.seed}\n"
                                                     f"recorded at {self.__simulator.time}.\n"
                                                     f"Press {ColorConfig.highlight_key(__space)} to execute the "
                                                     "simulation step by step. Alternatively, if you keep it pressed "
                                                     "the simulation will be executed automatically with short delays "
                                                     "after each step until you let go of Space again.")
            else:
                Popup.message("Simulating other version", "You try to simulate the run of a different game version.\n"
                                                          f"Your version: {Config.version()}\n"
                                                          f"Version you try to simulate: {self.__simulator.version}\n"
                                                          "This is not supported and can cause problems. Only "
                                                          "continue if you know what you do! Else close this popup "
                                                          "and press ESC to abort the simulation.")
        except FileNotFoundError:
            Logger.instance().show_error(f"File \"{path}\" could not be found!")

    def _ready_for_input(self, key_pressed: int, gameplay: bool = True) -> bool:
        if self.__last_key != key_pressed:
            self.__last_key = key_pressed
            return True

        if gameplay:
            pause = GameplayConfig.gameplay_key_pause()
        else:
            pause = GameplayConfig.simulation_key_pause()
        now_time = time.time()
        if now_time - self.__last_input >= pause:
            self.__last_input = now_time
            return True
        return False

    def _handle_key_presses(self, key_pressed):
        if key_pressed == py_cui.keys.KEY_CTRL_Q:
            super(QrogueCUI, self)._handle_key_presses(py_cui.keys.KEY_ESCAPE)
        if self.__simulator is None:
            if self._ready_for_input(key_pressed, gameplay=True):
                if key_pressed == py_cui.keys.KEY_ESCAPE:
                    pass    # ignore ESC because this makes you leave the CUI
                else:
                    if GameplayConfig.log_keys() and self.__game_started and not self.is_simulating:
                        self.__key_logger.log(self.__controls, key_pressed)
                    super(QrogueCUI, self)._handle_key_presses(key_pressed)
        elif key_pressed in self.__controls.get_keys(Keys.StopSimulator):
            Popup.message("Simulator", "stopped Simulator")
            self.__simulator = None
        else:
            if self._ready_for_input(key_pressed, gameplay=False):
                key = self.__simulator.next()
                if key:
                    super(QrogueCUI, self)._handle_key_presses(key)
                else:
                    Popup.message("Simulator", "finished")
                    self.__simulator = None

    def _draw(self, stdscr) -> None:    # overridden because we want to ignore mouse events
        """Main CUI draw loop called by start()

        Parameters
        --------
        stdscr : curses Standard screen
            The screen buffer used for drawing CUI elements
        """

        self._stdscr = stdscr
        key_pressed = 0

        # Clear and refresh the screen for a blank canvas
        stdscr.clear()
        stdscr.refresh()
        #curses.mousemask(curses.ALL_MOUSE_EVENTS)   # ignore mouse events for our CUI
        # stdscr.nodelay(False)
        #stdscr.keypad(True)

        # Initialization functions. Generates colors and renderer
        self._initialize_colors()
        self._initialize_widget_renderer()

        # If user specified a refresh timeout, apply it here
        if self._refresh_timeout > 0:
            self._stdscr.timeout(self._refresh_timeout)

        # If user sets non-default border characters, update them here
        if self._border_characters is not None and self._renderer is not None:
            self._renderer._set_border_renderer_chars(self._border_characters)

        # Loop where key_pressed is the last character pressed. Wait for exit key while no popup or focus mode
        while key_pressed != self._exit_key or self._in_focused_mode or self._popup is not None:

            try:
                # If we call stop, we want to break out of the main draw loop
                if self._stopped:
                    break

                # Initialization and size adjustment
                stdscr.erase()

                # If the user defined an update function to fire on each draw call,
                # Run it here. This can of course be also handled user-side
                # through a separate thread.
                if self._on_draw_update_func is not None:
                    self._on_draw_update_func()

                # This is what allows the CUI to be responsive. Adjust grid size based on current terminal size
                # Resize the grid and the widgets if there was a resize operation
                if key_pressed == curses.KEY_RESIZE:
                    try:
                        self._refresh_height_width()
                    except py_cui.errors.PyCUIOutOfBoundsError as e:
                        self._logger.info('Resized terminal too small')
                        self._display_window_warning(stdscr, str(e))

                # Here we handle mouse click events globally, or pass them to the UI element to handle
                elif key_pressed == curses.KEY_MOUSE:
                    self._logger.info('Detected mouse click')

                    valid_mouse_event = True
                    try:
                        id, x, y, _, mouse_event = curses.getmouse()
                    except curses.error as e:
                        valid_mouse_event = False
                        self._logger.error(f'Failed to handle mouse event: {str(e)}')

                    if valid_mouse_event:
                        in_element = self.get_element_at_position(x, y)

                        # In first case, we click inside already selected widget, pass click for processing
                        if in_element is not None:
                            self._logger.info(f'handling mouse press for elem: {in_element.get_title()}')
                            in_element._handle_mouse_press(x, y, mouse_event)

                        # Otherwise, if not a popup, select the clicked on widget
                        elif in_element is not None and not isinstance(in_element, py_cui.popups.Popup):
                            self.move_focus(in_element)
                            in_element._handle_mouse_press(x, y, mouse_event)

                # If we have a post_loading_callback, fire it here
                if self._post_loading_callback is not None and not self._loading:
                    self._logger.debug(f'Firing post-loading callback function {self._post_loading_callback.__name__}')
                    self._post_loading_callback()
                    self._post_loading_callback = None

                # Handle widget cycling
                if key_pressed == self._forward_cycle_key:
                    self._cycle_widgets()
                elif key_pressed == self._reverse_cycle_key:
                    self._cycle_widgets(reverse=True)

                # Handle keypresses
                self._handle_key_presses(key_pressed)

                try:
                    # Draw status/title bar, and all widgets. Selected widget will be bolded.
                    self._draw_status_bars(stdscr, self._height, self._width)
                    self._draw_widgets()
                    # draw the popup if required
                    if self._popup is not None:
                        self._popup._draw()

                    # If we are in live debug mode, we draw our debug messages
                    if self._logger.is_live_debug_enabled():
                        self._logger.draw_live_debug()

                except curses.error as e:
                    self._logger.error('Curses error while drawing TUI')
                    self._display_window_warning(stdscr, str(e))
                except py_cui.errors.PyCUIOutOfBoundsError as e:
                    self._logger.error('Resized terminal too small')
                    self._display_window_warning(stdscr, str(e))

                # Refresh the screen
                stdscr.refresh()

                # Wait for next input
                if self._loading or self._post_loading_callback is not None:
                    # When loading, refresh screen every quarter second
                    time.sleep(0.25)
                    # Need to reset key_pressed, because otherwise the previously pressed key will be used.
                    key_pressed = 0
                elif self._stopped:
                    key_pressed = self._exit_key
                else:
                    self._logger.info('Waiting for next keypress')
                    key_pressed = stdscr.getch()

            except KeyboardInterrupt:
                self._logger.info('Detect Keyboard Interrupt, Exiting...')
                self._stopped = True


        stdscr.erase()
        stdscr.refresh()
        curses.endwin()
        if self._on_stop is not None:
            self._logger.debug(f'Firing onstop function {self._on_stop.__name__}')
            self._on_stop()

    def _initialize_widget_renderer(self):
        """Function that creates the renderer object that will draw each widget
        """
        if self._renderer is None:
            self._renderer = MultiColorRenderer(self, self._stdscr, self._logger)
        super(QrogueCUI, self)._initialize_widget_renderer()

    def __init_keys(self) -> None:
        # debugging stuff
        self.add_key_command(self.__controls.get_key(Keys.PrintScreen), self.print_screen)
        self.__menu.get_main_widget().add_key_command(self.__controls.get_keys(Keys.PrintScreen), self.print_screen)
        self.__explore.get_main_widget().add_key_command(self.__controls.get_keys(Keys.PrintScreen), self.print_screen)
        self.__fight.get_main_widget().add_key_command(self.__controls.get_keys(Keys.PrintScreen), self.print_screen)
        self.__boss_fight.get_main_widget().add_key_command(self.__controls.get_keys(Keys.PrintScreen),
                                                            self.print_screen)

        self.__pause.get_main_widget().add_key_command(self.__controls.get_keys(Keys.CheatInput),
                                                       CheatConfig.cheat_input)
        self.__pause.get_main_widget().add_key_command(self.__controls.get_keys(Keys.CheatList), CheatConfig.cheat_list)
        # don't add the pause key to Menu and Pause itself!
        for widget_set in [self.__spaceship, self.__navigation, self.__explore, self.__fight, self.__boss_fight,
                           self.__shop, self.__riddle]:
            for widget in widget_set.get_widget_list():
                widget.widget.add_key_command(self.__controls.get_keys(Keys.Pause), Pausing.pause)

        # all selections
        selection_widgets = [
            self.__menu.selection,
            self.__workbench.selection, self.__workbench.upgrades,
            self.__fight.choices, self.__fight.details,
            self.__boss_fight.choices, self.__boss_fight.details,
            self.__shop.inventory, self.__shop.buy,
            self.__riddle.choices, self.__riddle.details,
            self.__pause.choices, self.__pause.details,
        ]
        for my_widget in selection_widgets:
            widget = my_widget.widget
            widget.add_key_command(self.__controls.get_keys(Keys.SelectionUp), my_widget.up)
            widget.add_key_command(self.__controls.get_keys(Keys.SelectionRight), my_widget.right)
            widget.add_key_command(self.__controls.get_keys(Keys.SelectionDown), my_widget.down)
            widget.add_key_command(self.__controls.get_keys(Keys.SelectionLeft), my_widget.left)

        # menu
        self.__menu.selection.widget.add_key_command(self.__controls.action, self.__use_menu_selection)

        # spaceship
        w = self.__spaceship.get_main_widget()
        w.add_key_command(self.__controls.get_keys(Keys.MoveUp), self.__spaceship.move_up)
        w.add_key_command(self.__controls.get_keys(Keys.MoveRight), self.__spaceship.move_right)
        w.add_key_command(self.__controls.get_keys(Keys.MoveDown), self.__spaceship.move_down)
        w.add_key_command(self.__controls.get_keys(Keys.MoveLeft), self.__spaceship.move_left)

        # workbench
        self.__workbench.selection.widget.add_key_command(self.__controls.action, self.__workbench_selection)
        self.__workbench.upgrades.widget.add_key_command(self.__controls.action, self.__workbench_upgrades)

        # navigation
        w = self.__navigation.get_main_widget()
        w.add_key_command(self.__controls.get_keys(Keys.MoveUp), self.__navigation.move_up)
        w.add_key_command(self.__controls.get_keys(Keys.MoveRight), self.__navigation.move_right)
        w.add_key_command(self.__controls.get_keys(Keys.MoveDown), self.__navigation.move_down)
        w.add_key_command(self.__controls.get_keys(Keys.MoveLeft), self.__navigation.move_left)

        # pause
        self.__pause.choices.widget.add_key_command(self.__controls.action, self.__pause_choices)
        self.__pause.details.widget.add_key_command(self.__controls.action, self.__pause_details)

        # explore
        w = self.__explore.get_main_widget()
        w.add_key_command(self.__controls.get_keys(Keys.MoveUp), self.__explore.move_up)
        w.add_key_command(self.__controls.get_keys(Keys.MoveRight), self.__explore.move_right)
        w.add_key_command(self.__controls.get_keys(Keys.MoveDown), self.__explore.move_down)
        w.add_key_command(self.__controls.get_keys(Keys.MoveLeft), self.__explore.move_left)

        # fight
        self.__fight.choices.widget.add_key_command(self.__controls.action, self.__fight_choices)
        self.__fight.details.widget.add_key_command(self.__controls.action, self.__fight_details)
        self.__fight.details.widget.add_key_command(self.__controls.get_keys(Keys.Cancel), self.__fight_details_back)
        self.__boss_fight.choices.widget.add_key_command(self.__controls.action, self.__boss_fight_choices)
        self.__boss_fight.details.widget.add_key_command(self.__controls.action, self.__boss_fight_details)
        self.__boss_fight.details.widget.add_key_command(self.__controls.get_keys(Keys.Cancel),
                                                         self.__boss_fight_details_back)

        # shop
        self.__shop.inventory.widget.add_key_command(self.__controls.action, self.__shop_inventory)
        self.__shop.buy.widget.add_key_command(self.__controls.action, self.__shop_buy)

        # riddle
        self.__riddle.choices.widget.add_key_command(self.__controls.action, self.__riddle_choices)
        self.__riddle.details.widget.add_key_command(self.__controls.action, self.__riddle_details)
        self.__riddle.details.widget.add_key_command(self.__controls.get_keys(Keys.Cancel), self.__riddle_details_back)

    def print_screen(self) -> None:
        text = ""
        for my_widget in self.__cur_widget_set.get_widget_list():
            text += str(my_widget) + "\n"
            text += my_widget.widget.get_title()
            text += "\n"
        PathConfig.new_screen_print(text)

    def apply_widget_set(self, new_widget_set: MyWidgetSet) -> None:
        new_widget_set.reset()
        super().apply_widget_set(new_widget_set)
        self.__cur_widget_set = new_widget_set
        self.move_focus(self.__cur_widget_set.get_main_widget(), auto_press_buttons=False)
        self.__cur_widget_set.render()

    def __show_message_popup(self, title: str, text: str, color: int) -> None:
        self.__focused_widget = self.get_selected_widget()
        self._popup = MultilinePopup(self, title, text, color, self._renderer, self._logger, self.__controls)

    def __show_input_popup(self, title: str, color: int, callback: Callable[[str], None]) -> None:
        self.__focused_widget = self.get_selected_widget()
        self._popup = py_cui.popups.TextBoxPopup(self, title, color, callback, self._renderer, False, self._logger)

    def __show_confirmation_popup(self, title: str, text: str, color: int, callback: Callable[[bool], None]):
        self.__focused_widget = self.get_selected_widget()
        self._popup = MultilinePopup(self, title, text, color, self._renderer, self._logger, self.__controls, callback)

    def close_popup(self) -> None:
        super(QrogueCUI, self).close_popup()
        self.move_focus(self.__focused_widget)
        Popup.on_close()

    def __dummy(self) -> None:
        pass

    def __general_continue(self):
        self.__state_machine.change_state(self.__state_machine.prev_state, None)

    def __start_playing(self):
        self.__state_machine.change_state(State.Spaceship, SaveData.instance())

    def switch_to_spaceship(self, data=None):
        # todo maybe data is the save data?
        if data:
            self.__spaceship.set_data(self.__spaceship_map)
        self.apply_widget_set(self.__spaceship)

    def __continue_spaceship(self) -> None:
        self.__state_machine.change_state(State.Spaceship, None)
        print("continue spaceship")

    def __use_workbench(self, direction: Direction, controllable: Controllable):
        self.__state_machine.change_state(State.Workbench, SaveData.instance())

    def switch_to_workbench(self, data=None):
        self.apply_widget_set(self.__workbench)

    def switch_to_menu(self, data=None) -> None:
        self.__menu.new_seed()
        self.apply_widget_set(self.__menu)

    def __show_world(self, world: WorldMap = None) -> None:
        if world is None:
            self.__state_machine.change_state(State.Spaceship, None)
        else:
            self.__state_machine.change_state(State.Navigation, world)

    def switch_to_navigation(self, data) -> None:
        if data is not None:
            self.__navigation.set_data(data)
        self.apply_widget_set(self.__navigation)

    def __start_level(self, seed: int, level: LevelMap) -> None:
        SaveData.instance().achievement_manager.reset_level_events()
        robot = level.controllable_tile.controllable
        if isinstance(robot, Robot):
            self.__pause.set_data(robot, level.name)   # needed for the HUD
            self.__state_machine.change_state(State.Explore, level)

            self.__key_logger = KeyLogger(seed)     # the seed used to build the Map
            self.__game_started = True
        else:
            Logger.instance().throw(ValueError(f"Tried to start a level with a non-Robot: {robot}"))

    def __game_over(self) -> None:
        Popup.message("Game Over!", "You Robot was out of energy so your mission failed. You will return to the "
                                    "Spaceship now.")
        self.__state_machine.change_state(State.Spaceship, None)

    def __won_tutorial(self) -> None:
        self.switch_to_menu(None)
        bell = ColorConfig.highlight_word("Bell")
        Popup.message("You won!", f"Congratulations, you defeated {bell} and successfully played the Tutorial!")

    def __start_fight(self, robot: Robot, enemy: Enemy, direction: Direction) -> None:
        self.__state_machine.change_state(State.Fight, (robot, enemy))

    def __start_boss_fight(self, robot: Robot, boss: Boss, direction: Direction):
        self.__state_machine.change_state(State.BossFight, (robot, boss))

    def switch_to_pause(self, data=None) -> None:
        self.apply_widget_set(self.__pause)

    def __pause_game(self) -> None:
        self.__state_machine.change_state(State.Pause, None)
        if not SaveData.instance().achievement_manager.check_achievement(achievements.EnteredPauseMenu):
            Popup.message("Pause", HelpText.get(HelpTextType.Pause))
            SaveData.instance().achievement_manager.finished_tutorial(achievements.EnteredPauseMenu)

    def switch_to_explore(self, data) -> None:
        if data is not None:
            map = data
            self.__explore.set_data(map)
        self.apply_widget_set(self.__explore)
        if data is not None:
            # we cannot do this in the same if because we need to apply the widget set first otherwise the focus will
            # be on the wrong widget after closing the popup
            if not SaveData.instance().achievement_manager.check_achievement(achievements.FinishedTutorial):
                Popup.scientist_says(HelpText.get(HelpTextType.FirstLevelIntroduction))

    def __continue_explore(self) -> None:
        self.__state_machine.change_state(State.Explore, None)

    def switch_to_fight(self, data) -> None:
        if data is not None:
            robot = data[0]
            enemy = data[1]
            self.__fight.set_data(robot, enemy)
        self.apply_widget_set(self.__fight)

    def switch_to_boss_fight(self, data) -> None:
        if data is not None:
            player = data[0]
            boss = data[1]
            self.__boss_fight.set_data(player, boss)
        self.apply_widget_set(self.__boss_fight)

    def __open_riddle(self, robot: Robot, riddle: Riddle):
        self.__state_machine.change_state(State.Riddle, (robot, riddle))

    def switch_to_riddle(self, data) -> None:
        if data is not None:
            player = data[0]
            riddle = data[1]
            self.__riddle.set_data(player, riddle)
        self.apply_widget_set(self.__riddle)

    def __visit_shop(self, robot: Robot, items: "list of ShopItems"):
        self.__state_machine.change_state(State.Shop, (robot, items))

    def switch_to_shop(self, data) -> None:
        if data is not None:
            player = data[0]
            items = data[1]
            self.__shop.set_data(player, items)
        self.apply_widget_set(self.__shop)

    def render(self) -> None:
        self.__render([self.__cur_widget_set])

    def __render(self, renderables: List[Renderable]):
        for r in renderables:
            r.render()

    def __use_menu_selection(self) -> None:
        if self.__menu.selection.use() and self.__cur_widget_set is self.__menu:
            self.render()

    def __workbench_selection(self) -> None:
        if self.__workbench.selection.use() and self.__cur_widget_set is self.__workbench:
            self.move_focus(self.__workbench.upgrades.widget, auto_press_buttons=False)
            self.render()

    def __workbench_upgrades(self) -> None:
        if self.__workbench.upgrades.use() and self.__cur_widget_set is self.__workbench:
            self.move_focus(self.__workbench.selection.widget, auto_press_buttons=False)
            self.render()

    def __pause_choices(self) -> None:
        if self.__pause.choices.use() and self.__cur_widget_set is self.__pause:
            self.move_focus(self.__pause.details.widget, auto_press_buttons=False)
            self.__render([self.__pause.choices, self.__pause.details])

    def __pause_details(self) -> None:
        if self.__pause.details.use() and self.__cur_widget_set is self.__pause:
            self.move_focus(self.__pause.choices.widget, auto_press_buttons=False)
            self.__pause.details.render_reset()
            self.render()

    def __fight_choices(self) -> None:
        if self.__fight.choices.use() and self.__cur_widget_set is self.__fight:
            self.move_focus(self.__fight.details.widget, auto_press_buttons=False)
            self.__render([self.__fight.choices, self.__fight.details])

    def __fight_details(self) -> None:
        if self.__fight.details.use():
            self.__fight_details_back()

    def __fight_details_back(self) -> None:
        if self.__cur_widget_set is self.__fight:
            self.move_focus(self.__fight.choices.widget, auto_press_buttons=False)
            self.__fight.choices.validate_index()   # somehow it can happen that the index is out of bounds after
                                                    # coming back from details which is why we validate it now
            self.__fight.details.render_reset()
            self.render()   # render the whole widget_set for updating the StateVectors and the circuit

    def __boss_fight_choices(self) -> None:
        if self.__boss_fight.choices.use() and self.__cur_widget_set is self.__boss_fight:
            self.move_focus(self.__boss_fight.details.widget, auto_press_buttons=False)
            self.__render([self.__boss_fight.choices, self.__boss_fight.details])

    def __boss_fight_details_back(self) -> None:
        if self.__cur_widget_set is self.__boss_fight:
            self.move_focus(self.__boss_fight.choices.widget, auto_press_buttons=False)
            self.__boss_fight.choices.validate_index()  # somehow it can happen that the index is out of bounds after
            # coming back from details which is why we validate it now
            self.__boss_fight.details.render_reset()
            self.render()  # render the whole widget_set for updating the StateVectors and the circuit

    def __boss_fight_details(self) -> None:
        if self.__boss_fight.details.use():
            self.__boss_fight_details_back()

    def __riddle_choices(self):
        if self.__riddle.choices.use() and self.__cur_widget_set is self.__riddle:
            self.move_focus(self.__riddle.details.widget, auto_press_buttons=False)
            self.__render([self.__riddle.choices, self.__riddle.details])

    def __riddle_details_back(self) -> None:
        if self.__cur_widget_set is self.__riddle:
            self.move_focus(self.__riddle.choices.widget, auto_press_buttons=False)
            self.__riddle.choices.validate_index()   # somehow it can happen that the index is out of bounds after
                                                    # coming back from details which is why we validate it now
            self.__riddle.details.render_reset()
            self.render()   # render the whole widget_set for updating the StateVectors and the circuit

    def __riddle_details(self) -> None:
        if self.__riddle.details.use():
            self.__riddle_details_back()

    def __shop_inventory(self) -> None:
        if self.__shop.inventory.use() and self.__cur_widget_set is self.__shop:
            self.move_focus(self.__shop.buy.widget, auto_press_buttons=False)
            self.render()

    def __shop_buy(self) -> None:
        if self.__shop.buy.use() and self.__cur_widget_set is self.__shop:
            self.move_focus(self.__shop.inventory.widget, auto_press_buttons=False)
            self.__shop.inventory.validate_index()   # somehow it can happen that the index is out of bounds after
                                                     # coming back from details which is why we validate it now
            self.__shop.details.render_reset()
            self.__shop.buy.render_reset()
            #self.__shop.buy.clear_text() # not needed since render_reset does this for second SelectionWidgets
            self.render()


class State(Enum):
    Menu = 0
    Pause = 1
    Explore = 2
    Fight = 3
    Shop = 4
    Riddle = 5
    BossFight = 6

    Spaceship = 7
    Workbench = 8
    Navigation = 9


class StateMachine:
    def __init__(self, renderer: QrogueCUI):
        self.__renderer = renderer
        self.__cur_state = None
        self.__prev_state = None

    @property
    def cur_state(self) -> State:
        return self.__cur_state

    @property
    def prev_state(self) -> State:
        return self.__prev_state

    def change_state(self, state: State, data) -> None:
        self.__prev_state = self.__cur_state
        self.__cur_state = state

        if self.__cur_state == State.Menu:
            self.__renderer.switch_to_menu(data)
        elif self.__cur_state == State.Explore:
            self.__renderer.switch_to_explore(data)
        elif self.__cur_state == State.Fight:
            self.__renderer.switch_to_fight(data)
        elif self.__cur_state == State.Riddle:
            self.__renderer.switch_to_riddle(data)
        elif self.__cur_state == State.Shop:
            self.__renderer.switch_to_shop(data)
        elif self.__cur_state == State.BossFight:
            self.__renderer.switch_to_boss_fight(data)
        elif self.__cur_state == State.Pause:
            self.__renderer.switch_to_pause()

        elif self.__cur_state == State.Spaceship:
            self.__renderer.switch_to_spaceship(data)
        elif self.__cur_state == State.Workbench:
            self.__renderer.switch_to_workbench(data)
        elif self.__cur_state == State.Navigation:
            self.__renderer.switch_to_navigation(data)



from abc import ABC, abstractmethod
from typing import List, Any, Callable, Tuple, Optional

from py_cui.widgets import BlockLabel

from qrogue.game.logic.actors import Robot
from qrogue.game.world.map import Map
from qrogue.game.world.navigation import Direction
from qrogue.util import ColorConfig, Controls, InstructionConfig, Keys, Logger, util_functions as uf, Config

from qrogue.graphics.widgets import Renderable


class MyBaseWidget(BlockLabel):
    def __init__(self, wid, title, grid, row, column, row_span, column_span, padx, pady, center, logger):
        super().__init__(wid, title, grid, row, column, row_span, column_span, padx, pady, center, logger)

    def set_title(self, title: str) -> None:
        super(MyBaseWidget, self).set_title(title)

    def get_title(self) -> str:
        return super(MyBaseWidget, self).get_title()

    def add_text_color_rule(self, regex: str, color: int, rule_type: str, match_type: str = 'line',
                            region: List[int] = [0,1], include_whitespace: bool=False, selected_color=None) -> None:
        super(MyBaseWidget, self).add_text_color_rule(regex, color, rule_type, match_type, region, include_whitespace,
                                                      selected_color)

    def add_key_command(self, keys: List[int], command: Callable[[],Any]) -> Any:
        for key in keys:
            super(MyBaseWidget, self).add_key_command(key, command)


class Widget(Renderable, ABC):
    def __init__(self, widget: MyBaseWidget):
        self.__widget = widget

    @property
    def widget(self) -> MyBaseWidget:
        return self.__widget

    @abstractmethod
    def set_data(self, data) -> None:
        pass

    @abstractmethod
    def render(self) -> None:
        pass

    @abstractmethod
    def render_reset(self) -> None:
        pass


class SimpleWidget(Widget):
    def __init__(self, widget: MyBaseWidget):
        super().__init__(widget)
        self.__text = ""

    def set_data(self, data) -> None:
        self.__text = str(data)

    def render(self) -> None:
        self.widget.set_title(self.__text)

    def render_reset(self) -> None:
        self.__text = ""
        self.widget.set_title("")


class HudWidget(Widget):
    def __init__(self, widget: MyBaseWidget):
        super().__init__(widget)
        self.__robot = None
        self.__map_name = None
        self.__render_duration = None

    def set_data(self, data: Tuple[Robot, Optional[str]]) -> None:
        self.__robot = data[0]
        if data[1]:
            self.__map_name = data[1]

    def reset_data(self) -> None:
        self.__robot = None

    def update_render_duration(self, duration: float):
        if Config.debugging():
            self.__render_duration = duration * 1000

    def render(self) -> None:
        text = ""
        if self.__map_name:
            text += f"{self.__map_name}\t"
        if self.__robot:
            text += f"Energy: {self.__robot.cur_energy} / {self.__robot.max_energy}   \t" \
                   f"{self.__robot.backpack.coin_count}$, {self.__robot.key_count()} keys"
        if self.__render_duration:
            text += f"\t\t{self.__render_duration:.2f} ms"
        self.widget.set_title(text)

    def render_reset(self) -> None:
        self.widget.set_title("")


class CircuitWidget(Widget):
    def __init__(self, widget: MyBaseWidget):
        super().__init__(widget)
        self.__robot = None
        # highlight everything between {} (gates), |> (start) or <| (end)
        widget.add_text_color_rule("(\{.*?\}|\|.*?\>|\<.*?\|)", ColorConfig.CIRCUIT_COLOR, 'contains',
                                   match_type='regex')

    def set_data(self, robot: Robot) -> None:
        self.__robot = robot

    def render(self) -> None:
        if self.__robot is not None:
            circ_str = self.__robot.get_circuit_print()
            self.widget.set_title(circ_str)

    def render_reset(self) -> None:
        self.widget.set_title("")


class MapWidget(Widget):
    def __init__(self, widget: MyBaseWidget):
        super().__init__(widget)
        self.__map = None
        self.__backup = None

    def set_data(self, map: Map) -> None:
        self.__map = map

    def render(self) -> None:
        if self.__map is not None:
            rows = self.__map.row_strings()
            # add robot
            x = self.__map.controllable_pos.x
            y = self.__map.controllable_pos.y
            rows[y] = rows[y][0:x] + self.__map.controllable_tile.get_img() + rows[y][x + 1:]

            self.widget.set_title("\n".join(rows))

    def render_reset(self) -> None:
        self.__backup = self.widget.get_title().title()
        self.widget.set_title("")

    def move(self, direction: Direction) -> bool:
        return self.__map.move(direction)


class TargetStateVectorWidget(Widget):
    def __init__(self, widget: MyBaseWidget, headline: str):
        super().__init__(widget)
        self.__headline = headline
        self.__state_vector_str = None
        widget.add_text_color_rule("~.*~", ColorConfig.STV_HEADING_COLOR, 'contains', match_type='regex')

    def set_data(self, state_vector_str: str) -> None:
        self.__state_vector_str = state_vector_str

    def render(self) -> None:
        if self.__state_vector_str:
            str_rep = f"~{self.__headline}~\n{self.__state_vector_str}"
            self.widget.set_title(str_rep)

    def render_reset(self) -> None:
        self.widget.set_title("")


class CurrentStateVectorWidget(Widget):
    def __init__(self, widget: MyBaseWidget, headline: str):
        super().__init__(widget)
        self.__headline = headline
        self.__state_vector_str = None
        self.__diff_vector_str = None
        widget.add_text_color_rule("~.*~", ColorConfig.STV_HEADING_COLOR, 'contains', match_type='regex')
        widget.add_text_color_rule("\(0\)", ColorConfig.CORRECT_AMPLITUDE_COLOR, "contains", match_type="regex")
        #widget.add_text_color_rule("\(\d\)", ColorConfig.CORRECT_AMPLITUDE_COLOR, "contains", match_type="regex")
        widget.add_text_color_rule("\([^0].*\)", ColorConfig.WRONG_AMPLITUDE_COLOR, "contains", match_type="regex")

    def set_data(self, state_vector_strings: Tuple[str, str]) -> None:
        self.__state_vector_str, self.__diff_vector_str = state_vector_strings

    def render(self) -> None:
        if self.__state_vector_str is not None:
            str_rep = f"~{self.__headline}~\n"
            stv_rows = self.__state_vector_str.split('\n')
            diff_rows = ["(" + val + ")" for val in self.__diff_vector_str.split('\n')]

            max_stv_width = max([len(val) for val in stv_rows])
            max_diff_width = max([len(val) for val in diff_rows])

            # last row is empty due to the trailing \n and therefore uninteresting to us
            for i in range(len(stv_rows) - 1):
                str_rep += uf.center_string(stv_rows[i], max_stv_width, uneven_left=True)
                str_rep += "  "
                str_rep += uf.center_string(diff_rows[i], max_diff_width, uneven_left=False)
                str_rep += "\n"
            self.widget.set_title(str_rep)

    def render_reset(self) -> None:
        self.widget.set_title("")


class QubitInfoWidget(Widget):
    def __init__(self, widget: MyBaseWidget, left_aligned: bool = True):
        super(QubitInfoWidget, self).__init__(widget)
        self.__left_aligned = left_aligned
        self.__text = ""
        widget.add_text_color_rule("~.*~", ColorConfig.QUBIT_INFO_COLOR, 'contains', match_type='regex')

    def set_data(self, num_of_qubits: int) -> None:
        head = ""
        body = ""
        box_left = "|" + " " * 2
        box_right = " " * 2 + "|"
        if self.__left_aligned:
            head_range = range(num_of_qubits-1, -1, -1)
        else:
            head_range = range(num_of_qubits)

        for i in head_range:
            head += f" q{i} "
        # skip 1 character left because of the additional ~ and 2 right because of the qubit index and ~
        head = box_left[:-1] + "~" + head[1:-1] + "~" + box_right[2:]

        for i in range(2 ** num_of_qubits):
            bin_num = bin(i)[2:]    # get rid of the '0b' at the beginning of the binary representation
            bin_num = bin_num.rjust(num_of_qubits, '0')     # add 0s to the beginning (left) by justifying the text to
                                                            # the right
            row = "   ".join(bin_num)  # separate the digits in the string with spaces
            if not self.__left_aligned:
                row = row[::-1] # [::-1] reverses the list so q0 is on the left
            body += box_left + row + box_right
            body += "\n"

        self.__text = head + "\n" + body
        self.widget.set_title(self.__text)

    def render(self) -> None:
        #self.widget.set_title(self.__text)
        pass

    def render_reset(self) -> None:
        self.widget.set_title("")


class SelectionWidget(Widget):
    __COLUMN_SEPARATOR = "   "

    @staticmethod
    def wrap_in_hotkey_str(options: List[str]) -> List[str]:
        if len(options) <= 1:
            return options      # no explicit hotkeys if there are not multiple options
        wrapped_options = []
        for i, option in enumerate(options):
            wrapped_options.append(f"[{i}] {option}")
        return wrapped_options

    def __init__(self, widget: MyBaseWidget, controls: Controls, columns: int = 1, is_second: bool = False,
                 stay_selected: bool = False):
        super(SelectionWidget, self).__init__(widget)
        self.__columns = columns
        self.__is_second = is_second
        self.__stay_selected = stay_selected
        self.__index = 0
        self.__choices = []
        self.__callbacks = []
        self.widget.add_text_color_rule(f"->", ColorConfig.SELECTION_COLOR, 'contains', match_type='regex')

        # sadly cannot use a loop here because of how lambda expressions work the index would be the same for all calls
        # instead we use a list of indices to still be flexible without changing much code
        indices = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        hotkeys = [
            controls.get_keys(Keys.HotKey0),
            controls.get_keys(Keys.HotKey1),
            controls.get_keys(Keys.HotKey2),
            controls.get_keys(Keys.HotKey3),
            controls.get_keys(Keys.HotKey4),
            controls.get_keys(Keys.HotKey5),
            controls.get_keys(Keys.HotKey6),
            controls.get_keys(Keys.HotKey7),
            controls.get_keys(Keys.HotKey8),
            controls.get_keys(Keys.HotKey9),
        ]
        self.widget.add_key_command(hotkeys[indices[0]], lambda : self.__jump_to_index(indices[0]))
        self.widget.add_key_command(hotkeys[indices[1]], lambda : self.__jump_to_index(indices[1]))
        self.widget.add_key_command(hotkeys[indices[2]], lambda : self.__jump_to_index(indices[2]))
        self.widget.add_key_command(hotkeys[indices[3]], lambda : self.__jump_to_index(indices[3]))
        self.widget.add_key_command(hotkeys[indices[4]], lambda : self.__jump_to_index(indices[4]))
        self.widget.add_key_command(hotkeys[indices[5]], lambda : self.__jump_to_index(indices[5]))
        self.widget.add_key_command(hotkeys[indices[6]], lambda : self.__jump_to_index(indices[6]))
        self.widget.add_key_command(hotkeys[indices[7]], lambda : self.__jump_to_index(indices[7]))
        self.widget.add_key_command(hotkeys[indices[8]], lambda : self.__jump_to_index(indices[8]))
        self.widget.add_key_command(hotkeys[indices[9]], lambda : self.__jump_to_index(indices[9]))

    @property
    def num_of_choices(self) -> int:
        return len(self.__choices)

    def update_text(self, text: str, index: int):
        if 0 <= index < len(self.__choices):
            self.__choices[index] = text

    def clear_text(self):
        self.__choices = []
        self.__callbacks = []

    def set_data(self, data: "tuple of list of str and list of SelectionCallbacks") -> None:
        self.render_reset()
        self.__choices, self.__callbacks = data
        choice_length = 0
        for choice in self.__choices:
            if len(choice) > choice_length:
                choice_length = len(choice)
        for i in range(len(self.__choices)):
            self.__choices[i] = self.__choices[i].ljust(choice_length)

    def render(self) -> None:
        str_rep = ""
        for i in range(self.num_of_choices):
            if i == self.__index and (self.widget.is_selected() or self.__stay_selected):
                wrapper = "-> "
            else:
                wrapper = "   "
            str_rep += wrapper
            str_rep += self.__choices[i]
            str_rep += " "
            if i % self.__columns == self.__columns - 1:
                str_rep += "\n"
            else:
                str_rep += self.__COLUMN_SEPARATOR
        self.widget.set_title(str_rep)

    def render_reset(self) -> None:
        self.widget.set_title("")
        self.__index = 0
        if self.__is_second:
            self.clear_text()

    def validate_index(self) -> bool:
        if self.__index < 0:
            self.__index = 0
            return False
        if len(self.__choices) <= self.__index:
            self.__index = len(self.__choices) - 1
            return False
        return True

    def __single_next(self) -> None:
        self.__index += 1
        if self.__index >= self.num_of_choices:
            self.__index = 0

    def __single_prev(self) -> None:
        self.__index -= 1
        if self.__index < 0:
            self.__index = self.num_of_choices - 1

    def up(self) -> None:
        if self.num_of_choices <= 1:
            return
        if self.num_of_choices <= self.__columns or self.__columns == 1:
            self.__single_prev()
        else:
            # special case for first line
            if self.__index < self.__columns:
                left_most = self.num_of_choices - self.num_of_choices % self.__columns
                self.__index = left_most + min(self.__index, self.num_of_choices % self.__columns - 1)
            else:
                self.__index -= self.__columns
        self.render()

    def right(self) -> None:
        if self.num_of_choices <= 1:
            return
        if self.__columns == 1 or self.num_of_choices <= self.__columns:
            self.__single_next()
        else:
            self.__index += 1
            if self.__index >= self.num_of_choices:
                self.__index -= (self.__index % self.__columns)
            elif self.__index % self.__columns == 0:
                self.__index -= self.__columns
        self.render()

    def down(self) -> None:
        if self.num_of_choices <= 1:
            return
        if self.num_of_choices <= self.__columns or self.__columns == 1:
            self.__single_next()
        else:
            # special case if we are currently in the last line
            if self.__index >= self.num_of_choices - (self.num_of_choices % self.__columns):
                self.__index = self.__index % self.__columns
            else:
                self.__index += self.__columns
                if self.__index >= self.num_of_choices:
                    self.__index = self.num_of_choices - 1
        self.render()

    def left(self) -> None:
        if self.num_of_choices <= 1:
            return
        if self.__columns == 1 or self.num_of_choices <= self.__columns:
            self.__single_prev()
        else:
            # special case if we are currently in the last line
            if self.__index >= self.num_of_choices - (self.num_of_choices % self.__columns):
                self.__index = self.num_of_choices - 1
            else:
                self.__index -= 1
                if self.__index < 0:
                    self.__index += self.__columns
        self.render()

    def __jump_to_index(self, index: int):
        if index < 0:
            self.__index = 0
        elif self.num_of_choices <= index:
            self.__index = self.num_of_choices - 1
        else:
            self.__index = index
        self.render()

    def use(self) -> bool:
        """
        :return: True if the focus should move, False if the focus should stay in this SelectionWidget
        """
        # if only one callback is given, it needs the index as parameter
        if len(self.__callbacks) == 1 and self.num_of_choices > 1:
            ret = self.__callbacks[0](self.__index)
        else:
            if self.__index >= len(self.__callbacks):
                Logger.instance().throw(IndexError(f"Invalid index = {self.__index} for {self.__callbacks}. "
                                                   f"Text of choices: {self.__choices}"))
            ret = self.__callbacks[self.__index]()
        if ret is None: # move focus if nothing is returned
            return True
        else:
            return ret



from abc import ABC, abstractmethod


class Renderable(ABC):
    __COUNTER = 0

    @staticmethod
    def count() -> str:
        """
        Utility function to find out about rendering order for debugging purposes
        :return: string representation of an globally incremented counter
        """
        Renderable.__COUNTER += 1
        return str(Renderable.__COUNTER)

    @abstractmethod
    def render(self) -> None:
        pass



from typing import List, Optional

from qrogue.game.logic.collectibles import Collectible, consumable, instruction as gates, pickup, ShopItem
from qrogue.util import Logger, MyRandom, RandomManager, ShopConfig


class CollectibleFactory:
    def __init__(self, pool: List[Collectible]) -> None:
        if pool is None or len(pool) < 1:
            Logger.instance().throw(f"invalid pool for CollectibleFactory: {pool}")
        self.__pool = pool.copy()
        self.__rm = RandomManager.create_new()
        self.__order_index = -1

    @property
    def pool_copy(self) -> List[Collectible]:
        return self.__pool.copy()

    def __produce(self, rm: Optional[MyRandom], remove: bool = False):
        if rm:
            return rm.get_element(self.__pool, remove=remove)
        else:
            self.__order_index += 1
            if self.__order_index >= len(self.__pool):
                self.__order_index = 0
            elem = self.__pool[self.__order_index]
            if remove:
                self.__pool.remove(elem)
            return elem

    def produce(self, rm: Optional[MyRandom]) -> Collectible:
        return self.__produce(rm, remove=False)

    def produce_multiple(self, rm: Optional[MyRandom], num_of_elements: int, unique_elements: bool = True) \
            -> List[Collectible]:
        if unique_elements:
            temp = self.pool_copy
            elements = []
            while len(elements) < num_of_elements:
                elem = self.__produce(rm, remove=True)
                elements.append(elem)
                if len(self.__pool) <= 0:
                    self.__pool = temp.copy()
            self.__pool = temp
            return elements
        else:
            return [self.__produce(rm, remove=False) for _ in range(num_of_elements)]


class OrderedCollectibleFactory(CollectibleFactory):
    """
    Same as CollectibleFactory but instead of defaulting to random production this
    class defaults to ordered production
    """
    @staticmethod
    def from_factory(factory: CollectibleFactory):
        return OrderedCollectibleFactory(factory.pool_copy)

    def __init__(self, pool: List[Collectible]):
        super().__init__(pool)

    def produce(self, rm: Optional[MyRandom]) -> Collectible:
        return super(OrderedCollectibleFactory, self).produce(None)

    def produce_multiple(self, rm: Optional[MyRandom], num_of_elements: int, unique_elements: bool = True) \
            -> List[Collectible]:
        return super(OrderedCollectibleFactory, self).produce_multiple(None, num_of_elements, unique_elements)


class GateFactory:
    @staticmethod
    def default():
        return CollectibleFactory(pool=[
            gates.HGate(), gates.XGate(), gates.YGate(), gates.ZGate(),
            gates.CXGate(), gates.SwapGate(),
        ])


class ShopFactory:
    @staticmethod
    def default() -> "ShopFactory":
        special_pool = [gates.HGate(), gates.XGate(), gates.HGate()]
        common_pool = [pickup.Key(1), pickup.Key(2), consumable.EnergyRefill(20)]
        return ShopFactory(common_pool, special_pool, quality_level=1)

    def __init__(self, common_pool: List[Collectible], special_pool: List[Collectible], quality_level: int = 1,
                 min_items: int = 2, max_items: int = 5, discount: bool = False):
        """

        :param common_pool: contains all Pickups the shop can sell
        :param special_pool: contains all Gates (and later other special Collectibles) that the Shop could offer
        :param quality_level: how good the Shop should be, i.e. higher quality = better ShopItems
        :param min_items: minimum number of ShopItems available in the Shop
        :param max_items: maximum number of ShopItems available in the Shop
        :param discount: whether the ShopItems are 50% off or not
        """
        self.__common_pool = common_pool
        self.__special_pool = special_pool
        self.__quality_level = quality_level
        self.__min_items = min_items
        self.__max_items = max_items
        self.__discount = discount

    def produce(self, rm: MyRandom, num_of_items: int = 0) -> List[ShopItem]:
        """
        Produces a random list of ShopItems based on the factory's parameter. Said parameter are not changed allowing
        to produce multiple lists with the same factory. The only thing that obviously changes is the current seed of
        its random manager.

        :param rm: decides about the randomness
        :param num_of_items: if set to 0 or less, takes a random number of items in the factory's [min, max[ range
        :return: a list of randomly picked ShopItems based on the factory's parameter
        """
        shop_items = []
        special_items = self.__special_pool.copy()
        pickups = self.__common_pool.copy()
        if num_of_items <= 0:
            num_of_items = rm.get_int(self.__min_items, self.__max_items)
        quality_level = self.__quality_level
        while len(shop_items) <= num_of_items:
            if quality_level > 0:
                # with a certain probability the next item will be a special one (1 is guaranteed at quality_level == 1)
                gets_special = rm.get() < 1.0 / quality_level
            else:
                gets_special = False

            if gets_special and self.__special_pool is not None and len(special_items) > 0:
                item = rm.get_element(special_items, remove=True)
            else:
                item = rm.get_element(pickups, remove=False)

            # calculate the price with possibly a small variation
            if self.__discount:
                price = round(item.default_price() / 2)
            else:
                price_sigma = max(ShopConfig.base_unit(), item.default_price() * 0.1)
                price = item.default_price() + round(rm.get(-price_sigma, +price_sigma))

            shop_items.append(ShopItem(item, price))
            quality_level -= 1
        return shop_items



# Generated from D:/Documents/pycharm_workspace/Qrogue/qrogue/game/world/dungeon_generator\QrogueDungeon.g4 by ANTLR 4.9.2
from antlr4 import *

# This class defines a complete listener for a parse tree produced by QrogueDungeonParser.
class QrogueDungeonListener(ParseTreeListener):

    # Enter a parse tree produced by QrogueDungeonParser#start.
    def enterStart(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#start.
    def exitStart(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#integer.
    def enterInteger(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#integer.
    def exitInteger(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#complex_number.
    def enterComplex_number(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#complex_number.
    def exitComplex_number(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#robot.
    def enterRobot(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#robot.
    def exitRobot(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#layout.
    def enterLayout(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#layout.
    def exitLayout(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#l_room_row.
    def enterL_room_row(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#l_room_row.
    def exitL_room_row(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#l_hallway_row.
    def enterL_hallway_row(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#l_hallway_row.
    def exitL_hallway_row(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#rooms.
    def enterRooms(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#rooms.
    def exitRooms(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#room.
    def enterRoom(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#room.
    def exitRoom(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_attributes.
    def enterR_attributes(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_attributes.
    def exitR_attributes(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_visibility.
    def enterR_visibility(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_visibility.
    def exitR_visibility(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_type.
    def enterR_type(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_type.
    def exitR_type(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#r_row.
    def enterR_row(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#r_row.
    def exitR_row(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#tile.
    def enterTile(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#tile.
    def exitTile(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#tile_descriptor.
    def enterTile_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#tile_descriptor.
    def exitTile_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#trigger_descriptor.
    def enterTrigger_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#trigger_descriptor.
    def exitTrigger_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#message_descriptor.
    def enterMessage_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#message_descriptor.
    def exitMessage_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#enemy_descriptor.
    def enterEnemy_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#enemy_descriptor.
    def exitEnemy_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#collectible_descriptor.
    def enterCollectible_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#collectible_descriptor.
    def exitCollectible_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#energy_descriptor.
    def enterEnergy_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#energy_descriptor.
    def exitEnergy_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#riddle_descriptor.
    def enterRiddle_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#riddle_descriptor.
    def exitRiddle_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#shop_descriptor.
    def enterShop_descriptor(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#shop_descriptor.
    def exitShop_descriptor(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#hallways.
    def enterHallways(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#hallways.
    def exitHallways(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#hallway.
    def enterHallway(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#hallway.
    def exitHallway(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#h_attributes.
    def enterH_attributes(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#h_attributes.
    def exitH_attributes(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#draw_strategy.
    def enterDraw_strategy(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#draw_strategy.
    def exitDraw_strategy(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stv_pools.
    def enterStv_pools(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stv_pools.
    def exitStv_pools(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#default_stv_pool.
    def enterDefault_stv_pool(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#default_stv_pool.
    def exitDefault_stv_pool(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stv_pool.
    def enterStv_pool(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stv_pool.
    def exitStv_pool(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stvs.
    def enterStvs(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stvs.
    def exitStvs(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#stv.
    def enterStv(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#stv.
    def exitStv(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#reward_pools.
    def enterReward_pools(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#reward_pools.
    def exitReward_pools(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#default_reward_pool.
    def enterDefault_reward_pool(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#default_reward_pool.
    def exitDefault_reward_pool(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#reward_pool.
    def enterReward_pool(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#reward_pool.
    def exitReward_pool(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#collectibles.
    def enterCollectibles(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#collectibles.
    def exitCollectibles(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#collectible.
    def enterCollectible(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#collectible.
    def exitCollectible(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#messages.
    def enterMessages(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#messages.
    def exitMessages(self, ctx):
        pass


    # Enter a parse tree produced by QrogueDungeonParser#message.
    def enterMessage(self, ctx):
        pass

    # Exit a parse tree produced by QrogueDungeonParser#message.
    def exitMessage(self, ctx):
        pass





# exporting
from .tiles import TileCode, TileColorer, Tile
from .tiles import Invalid, Debug, Void, Floor, Wall, Obstacle, FogOfWar, Decoration, ControllableTile
from .walk_trigger_tiles import WalkTriggerTile, Trigger, Teleport, Message, Riddler, ShopKeeper, Collectible, Energy
from .puzzle_tiles import Enemy, Boss
from .door_tiles import DoorOpenState, DoorOneWayState, Door, EntangledDoor, HallwayEntrance

# importing
# +util
# +game (target_factory)
# +logic
# +logic.actors
# +logic.actors.puzzles
# +logic.collectibles
# +navigation



from typing import Callable, List, Tuple, Dict

from antlr4 import InputStream, CommonTokenStream
from antlr4.tree.Tree import TerminalNodeImpl

from qrogue.game.logic import Message, StateVector
from qrogue.game.logic.actors import Controllable, Riddle
from qrogue.game.logic.actors.controllables import TestBot
from qrogue.game.logic.collectibles import Collectible, pickup, instruction, MultiCollectible, Qubit, ShopItem, \
    CollectibleFactory, OrderedCollectibleFactory
from qrogue.game.target_factory import EnemyFactory, ExplicitTargetDifficulty, TargetDifficulty
from qrogue.game.world import tiles
from qrogue.game.world.map import CallbackPack, LevelMap, rooms
from qrogue.game.world.navigation import Coordinate, Direction
from qrogue.util import Config, HelpText, MapConfig, MyRandom, PathConfig, Logger, CommonQuestions

from . import parser_util
from .generator import DungeonGenerator
from .dungeon_parser.QrogueDungeonLexer import QrogueDungeonLexer
from .dungeon_parser.QrogueDungeonParser import QrogueDungeonParser
from .dungeon_parser.QrogueDungeonVisitor import QrogueDungeonVisitor


class QrogueLevelGenerator(DungeonGenerator, QrogueDungeonVisitor):
    __DEFAULT_NUM_OF_SHOP_ITEMS = 3
    __DEFAULT_NUM_OF_RIDDLE_ATTEMPTS = 7
    __ROBOT_NO_GATES = "none"
    __SPAWN_ROOM_ID = "SR"

    @staticmethod
    def __normalize_reference(reference: str) -> str:
        return parser_util.normalize_reference(reference)

    @staticmethod
    def __tile_str_to_code(tile_str: str) -> tiles.TileCode:
        if tile_str.isdigit():
            return tiles.TileCode.Enemy

        elif tile_str == parser_util.COLLECTIBLE_TILE:
            return tiles.TileCode.Collectible
        elif tile_str == parser_util.TRIGGER_TILE:
            return tiles.TileCode.Trigger
        elif tile_str == parser_util.MESSAGE_TILE:
            return tiles.TileCode.Message
        elif tile_str == parser_util.ENERGY_TILE:
            return tiles.TileCode.Energy
        elif tile_str == parser_util.RIDDLER_TILE:
            return tiles.TileCode.Riddler
        elif tile_str == parser_util.SHOP_KEEPER_TILE:
            return tiles.TileCode.ShopKeeper
        elif tile_str == parser_util.FLOOR_TILE:
            return tiles.TileCode.Floor
        elif tile_str == parser_util.OBSTACLE_TILE:
            return tiles.TileCode.Obstacle
        else:
            return tiles.TileCode.Invalid

    @staticmethod
    def __tile_code_to_str(tile_code: tiles.TileCode) -> str:
        if tile_code is tiles.TileCode.Enemy:
            # todo print warning?
            return "0"
        elif tile_code is tiles.TileCode.Collectible:
            return parser_util.COLLECTIBLE_TILE
        elif tile_code is tiles.TileCode.Trigger:
            return parser_util.TRIGGER_TILE
        elif tile_code is tiles.TileCode.Message:
            return parser_util.MESSAGE_TILE
        elif tile_code is tiles.TileCode.Energy:
            return parser_util.ENERGY_TILE
        elif tile_code is tiles.TileCode.Riddler:
            return parser_util.RIDDLER_TILE
        elif tile_code is tiles.TileCode.ShopKeeper:
            return parser_util.SHOP_KEEPER_TILE
        elif tile_code is tiles.TileCode.Floor:
            return parser_util.FLOOR_TILE
        else:
            # todo print warning?
            return None

    def __init__(self, seed: int, check_achievement: Callable[[str], bool], trigger_event: Callable[[str], None],
                 load_map_callback: Callable[[str, Coordinate], None]):
        super().__init__(seed, 0, 0)
        self.__seed = seed
        self.__check_achievement = check_achievement
        self.__trigger_event = trigger_event
        self.__load_map = load_map_callback

        self.__warnings = 0
        self.__robot = None
        self.__rm = MyRandom(seed)

        self.__messages = {}        # str -> str

        self.__reward_pools = {}
        self.__default_reward_factory = None  # CollectibleFactory

        self.__stv_pools = {}           # str -> Tuple[List[StateVector], CollectibleFactory] (latter may be None)
        self.__default_target_difficulty = None  # ExplicitTargetDifficulty

        self.__default_enemy_factory = None     # needed to create default_tile enemies

        self.__hallways_by_id = {}      # hw_id -> Door
        self.__hallways = {}            # stores the

        #self.__template_events = PathConfig.read()
        self.__events = []

        self.__enemy_groups_by_room = {}    # room_id -> dic[1-9]
        self.__cur_room_id = None   # needed for enemy groups
        self.__rooms = {}
        self.__spawn_pos = None

        # holds references to already created hallways so that neighbors can use it instead of
        # creating their own, redundant hallway
        self.__created_hallways = {}

    @property
    def __cbp(self) -> CallbackPack:
        return CallbackPack.instance()

    def warning(self, text: str):
        parser_util.warning(text)
        self.__warnings += 1

    def generate(self, file_name: str, in_dungeon_folder: bool = True) -> Tuple[LevelMap, bool]:
        map_data = PathConfig.read_level(file_name, in_dungeon_folder)

        input_stream = InputStream(map_data)
        lexer = QrogueDungeonLexer(input_stream)
        token_stream = CommonTokenStream(lexer)
        parser = QrogueDungeonParser(token_stream)
        parser.addErrorListener(parser_util.MyErrorListener())

        try:
            name, room_matrix = self.visit(parser.start())
            if name is None:
                name = file_name
        except SyntaxError as se:
            Logger.instance().error(str(se))
            return None, False

        # add empty rooms if rows don't have the same width
        max_len = 0
        for row in room_matrix:
            if len(row) > max_len:
                max_len = len(row)
        for row in room_matrix:
            if len(row) < max_len:
                row += [None] * (max_len - len(row))

        level = LevelMap(name, file_name, self.__seed, room_matrix, self.__robot, self.__spawn_pos,
                         self.__check_achievement, self.__trigger_event)
        return level, True

    def _add_hallway(self, room1: Coordinate, room2: Coordinate, door: tiles.Door):
        if door:    # for simplicity door could be null so we check it here
            if room1 in self.__hallways:
                self.__hallways[room1][room2] = door
            else:
                self.__hallways[room1] = {room2: door}
            if room2 in self.__hallways:
                self.__hallways[room2][room1] = door
            else:
                self.__hallways[room2] = {room1: door}

    def __get_default_tile(self, tile_str: str, enemy_dic: Dict[int, List[tiles.Enemy]],
                           get_entangled_tiles: Callable[[int], List[tiles.Tile]]) -> tiles.Tile:
        if tile_str == parser_util.PLACE_HOLDER_TILE:
            return tiles.Floor()

        tile_code = QrogueLevelGenerator.__tile_str_to_code(tile_str)
        if tile_code is tiles.TileCode.Enemy:
            enemy_id = int(tile_str)
            enemy = tiles.Enemy(self.__default_enemy_factory, get_entangled_tiles, enemy_id)
            if enemy_id not in enemy_dic:
                enemy_dic[enemy_id] = []
            enemy_dic[enemy_id].append(enemy)
            return enemy

        elif tile_code is tiles.TileCode.Collectible:
            return tiles.Collectible(self.__default_reward_factory.produce(self.__rm))

        elif tile_code is tiles.TileCode.Trigger:
            return self.__load_trigger("*defaultTrigger")

        elif tile_code is tiles.TileCode.Message:
            # todo is it okay to print this?
            return tiles.Message.create("Hmm, I don't know what to say.", Config.scientist_name())

        elif tile_code is tiles.TileCode.Energy:
            return tiles.Collectible(pickup.Energy())

        elif tile_code is tiles.TileCode.Riddler:
            stv = self.__default_target_difficulty.create_statevector(self.__robot, self.__rm)
            reward = self.__default_reward_factory.produce(self.__rm)
            riddle = Riddle(stv, reward, self.__DEFAULT_NUM_OF_RIDDLE_ATTEMPTS)
            return tiles.Riddler(self.__cbp.open_riddle, riddle)

        elif tile_code is tiles.TileCode.ShopKeeper:
            items = self.__default_reward_factory.produce_multiple(self.__rm, self.__DEFAULT_NUM_OF_SHOP_ITEMS)
            return tiles.ShopKeeper(self.__cbp.visit_shop, [ShopItem(item) for item in items])

        elif tile_code is tiles.TileCode.Floor:
            return tiles.Floor()
        elif tile_code is tiles.TileCode.Obstacle:
            return tiles.Obstacle()
        else:
            self.warning(f"Unknown tile specified: {tile_str}. Using a Floor-Tile instead.")
            return tiles.Floor()

    def __get_draw_strategy(self, ctx: QrogueDungeonParser.Draw_strategyContext):
        if ctx:
            return self.visit(ctx)
        return False    # default value is "random"

    def __teleport_callback(self, map_name: str, spawn_pos: Coordinate):
        def cb(confirm: bool):
            if confirm:
                self.__load_map(map_name, spawn_pos)
        CommonQuestions.GoingBack.ask(cb)

    ##### load from references #####

    def __load_reward_pool(self, reference: str) -> CollectibleFactory:
        if reference in self.__reward_pools:
            return self.__reward_pools[reference]

        ref = self.__normalize_reference(reference)
        if ref in ['coin', 'coins']:
            pool = [pickup.Coin(1)]
        elif ref in ['key', 'keys']:
            pool = [pickup.Key(1)]
        elif ref in ['energy']:
            pool = [pickup.Energy()]
        else:
            self.warning(f"Imports not yet supported: {reference}")
            # todo load from somewhere else?
            pool = [pickup.Key(0)]
        return CollectibleFactory(pool)

    def __load_stv_pool(self, reference: str) -> TargetDifficulty:
        if reference in self.__stv_pools:
            return self.__stv_pools[reference]
        else:
            # todo implement imports

            return self.__default_target_difficulty

    def __load_gate(self, reference: str) -> instruction.Instruction:
        ref = self.__normalize_reference(reference)
        if ref in ['x', 'xgate']:
            return instruction.XGate()
        elif ref in ['y', 'ygate']:
            return instruction.YGate()
        elif ref in ['z', 'zgate']:
            return instruction.ZGate()
        elif ref in ['h', 'hgate', 'hadamard', 'hadamarggate']:
            return instruction.HGate()
        elif ref in ['cx', 'cxgate']:
            return instruction.CXGate()
        elif ref in ['swap', 'swapgate']:
            return instruction.SwapGate()

        elif ref not in ['i', 'igate']:
            self.warning(f"Unknown gate reference: {reference}. Returning I Gate instead.")
        return instruction.IGate()

    def __load_trigger(self, reference: str) -> tiles.Trigger:
        if reference.startswith("tp"):
            # teleport trigger
            ref = reference[2:]
            if ref.startswith("l"):
                level = ref[1:]
            elif ref.startswith("w"):
                world = ref[1:]

        # todo implement
        def callback(direction: Direction, controllable: Controllable):
            #Popup.message("Trigger", str(reference))
            pass

        return tiles.Trigger(callback)

    def __load_message(self, reference: str) -> Message:
        if reference in self.__messages:
            return self.__messages[reference]
        norm_ref = self.__normalize_reference(reference)
        if norm_ref in self.__messages:
            return self.__messages[norm_ref]
        elif norm_ref.startswith("helptext"):
            help_text_type = norm_ref[len("helptext"):]
            help_text = HelpText.load(help_text_type)
            if help_text:
                return Message.create_simple(norm_ref, help_text)
        self.warning(f"Unknown text reference: {reference}. Returning \"Message not found!\"")
        return Message.error("Message not found!")

    def __load_hallway(self, reference: str) -> tiles.Door:
        if reference in self.__hallways_by_id:
            return self.__hallways_by_id[reference]
        elif reference == parser_util.EMPTY_HALLWAY_CODE:
            return None
        elif reference == parser_util.DEFAULT_HALLWAY_STR:
            return tiles.Door(Direction.North)
        else:
            # todo implement hallway imports
            return tiles.Door(Direction.North)

    def __load_room(self, reference: str, x: int, y: int) -> rooms.Room:
        hw_dic = parser_util.get_hallways(self.__created_hallways, self.__hallways, Coordinate(x, y))
        if reference in self.__rooms:
            room = self.__rooms[reference]
            if room.type is rooms.AreaType.SpawnRoom:
                if self.__spawn_pos:
                    self.warning("A second SpawnRoom was defined! Ignoring the first one "
                                 "and using this one as SpawnRoom.")
                self.__spawn_pos = Coordinate(x, y)
            return room.copy(hw_dic)
        elif reference == self.__SPAWN_ROOM_ID:
            room = rooms.SpawnRoom(self.__load_map, None, hw_dic[Direction.North], hw_dic[Direction.East],
                                   hw_dic[Direction.South], hw_dic[Direction.West])
            if self.__spawn_pos:
                self.warning("A second SpawnRoom was defined! Ignoring the first one and using this one as "
                             "SpawnRoom.")
            self.__spawn_pos = Coordinate(x, y)
            return room
        elif reference[0] == '_':
            # todo handle templates
            room = rooms.CustomRoom(rooms.AreaType.Placeholder, None, hw_dic[Direction.North], hw_dic[Direction.East],
                                    hw_dic[Direction.South], hw_dic[Direction.West])
            self.__rooms[reference] = room
            return room
        else:
            self.warning(f"room_id \"{reference}\" not specified and imports not yet supported! "
                         "Placing an empty room instead.")
            room = rooms.Placeholder.empty_room(hw_dic[Direction.North], hw_dic[Direction.East],
                                                hw_dic[Direction.South], hw_dic[Direction.West])
            self.__rooms[reference] = room
            return room

    ##### General area

    def visitInteger(self, ctx: QrogueDungeonParser.IntegerContext) -> int:
        if ctx.DIGIT():
            return int(ctx.DIGIT().getText())
        elif ctx.HALLWAY_ID():
            return int(ctx.HALLWAY_ID().getText())
        elif ctx.INTEGER():
            return int(ctx.INTEGER().getText())
        else:
            return None

    def visitComplex_number(self, ctx: QrogueDungeonParser.Complex_numberContext) -> complex:
        if ctx.SIGN(0):
            if ctx.SIGN(0).symbol.type is QrogueDungeonParser.PLUS_SIGN:
                first_sign = "+"
            else:
                first_sign = "-"
        else:
            first_sign = "+"

        integer_ = ctx.integer()
        float_ = ctx.FLOAT()
        imag_ = ctx.IMAG_NUMBER()

        complex_number = first_sign
        if integer_ or float_:
            if integer_:
                num = str(self.visit(integer_))
            else:
                num = float_.getText()
            complex_number += num

            if ctx.SIGN(1):
                complex_number += ctx.SIGN(1).symbol.text + str(imag_)
        else:
            complex_number += str(imag_)

        return complex(complex_number)

    def visitDraw_strategy(self, ctx: QrogueDungeonParser.Draw_strategyContext) -> bool:
        """

        :param ctx:
        :return: True if strategy is 'ordered', False if it is 'random'
        """
        return ctx.ORDERED_DRAW() is not None

    ##### Message area ######

    def visitMessage(self, ctx: QrogueDungeonParser.MessageContext) -> Message:
        m_id = self.__normalize_reference(ctx.REFERENCE(0).getText())
        text = ctx.TEXT().getText()[1:-1]
        if ctx.EVENT_LITERAL():
            event = self.__normalize_reference(ctx.REFERENCE(1).getText())
            msg_ref = self.__normalize_reference(ctx.REFERENCE(2).getText())
            return Message(m_id, Config.scientist_name(), text, event, msg_ref)
        else:
            return Message.create_simple(m_id, text)

    def visitMessages(self, ctx: QrogueDungeonParser.MessagesContext):
        self.__messages.clear()
        for msg in ctx.message():
            message = self.visit(msg)
            self.__messages[message.id] = message
        for message in self.__messages.values():
            if message.alt_message_ref and message.alt_message_ref in self.__messages:
                alt_message = self.__messages[message.alt_message_ref]
                try:
                    message.resolve_message_ref(alt_message)
                except ValueError as ve:
                    self.warning(f"Message-cycle found: {ve}")

    ##### Reward Pool area #####

    def visitCollectible(self, ctx: QrogueDungeonParser.CollectibleContext) -> Collectible:
        if ctx.KEY_LITERAL():
            val = self.visit(ctx.integer())
            return pickup.Key(val)
        elif ctx.COIN_LITERAL():
            val = self.visit(ctx.integer())
            return pickup.Coin(val)
        elif ctx.ENERGY_LITERAL():
            val = self.visit(ctx.integer())
            return pickup.Energy(val)
        elif ctx.GATE_LITERAL():
            reference = ctx.REFERENCE().getText()
            return self.__load_gate(reference)
        elif ctx.QUBIT_LITERAL():
            if ctx.integer():
                return Qubit(self.visit(ctx.integer()))
            else:
                return Qubit(1)
        else:
            self.warning("No legal collectible specified!")
        return None

    def visitCollectibles(self, ctx: QrogueDungeonParser.CollectiblesContext) -> List[Collectible]:
        collectible_list = []
        for collectible in ctx.collectible():
            collectible_list.append(self.visit(collectible))
        return collectible_list

    def visitReward_pool(self, ctx: QrogueDungeonParser.Reward_poolContext) -> Tuple[str, CollectibleFactory]:
        pool_id = ctx.REFERENCE().getText()
        ordered = self.__get_draw_strategy(ctx.draw_strategy())
        collectible_list = self.visit(ctx.collectibles())
        if ordered:
            return pool_id, OrderedCollectibleFactory(collectible_list)
        else:
            return pool_id, CollectibleFactory(collectible_list)

    def visitDefault_reward_pool(self, ctx: QrogueDungeonParser.Default_reward_poolContext) -> CollectibleFactory:
        if ctx.REFERENCE():  # implicit definition
            pool_id = ctx.REFERENCE().getText()
            return self.__load_reward_pool(pool_id)

        else:  # explicit definition
            ordered = self.__get_draw_strategy(ctx.draw_strategy())
            collectible_list = self.visit(ctx.collectibles())
            if ordered:
                return OrderedCollectibleFactory(collectible_list)
            else:
                return CollectibleFactory(collectible_list)

    def visitReward_pools(self, ctx: QrogueDungeonParser.Reward_poolsContext) -> None:
        for reward_pool in ctx.reward_pool():
            pool_id, collectible_factory = self.visit(reward_pool)
            self.__reward_pools[pool_id] = collectible_factory
        self.__default_reward_factory = self.visit(ctx.default_reward_pool())

    ##### StateVector Pool area #####

    def visitStv(self, ctx: QrogueDungeonParser.StvContext) -> StateVector:
        amplitudes = []
        for cn in ctx.complex_number():
            amplitudes.append(self.visit(cn))
        if not StateVector.check_amplitudes(amplitudes):
            self.warning(f"Invalid amplitudes for StateVector: {amplitudes}! Using 0-only basis state instead.")
            amplitudes = [1] + [0] * (2 ** self.__robot.num_of_qubits - 1)
        return StateVector(amplitudes)

    def visitStvs(self, ctx: QrogueDungeonParser.StvsContext) -> List[StateVector]:
        stvs = []
        for stv in ctx.stv():
            stvs.append(self.visit(stv))
        return stvs

    def visitStv_pool(self, ctx: QrogueDungeonParser.Stv_poolContext) -> Tuple[str, TargetDifficulty]:
        pool_id = ctx.REFERENCE(0).getText()
        ordered = self.__get_draw_strategy(ctx.draw_strategy())
        stvs = self.visit(ctx.stvs())

        if ctx.REFERENCE(1):
            reward_pool_id = ctx.REFERENCE(1).getText()
            reward_factory = self.__load_reward_pool(reward_pool_id)
        else:
            reward_factory = self.__default_reward_factory

        return pool_id, ExplicitTargetDifficulty(stvs, reward_factory, ordered)

    def visitDefault_stv_pool(self, ctx: QrogueDungeonParser.Default_stv_poolContext) -> TargetDifficulty:
        if ctx.REFERENCE():  # implicit definition
            pool_id = ctx.REFERENCE().getText()
            return self.__load_stv_pool(pool_id)

        else:  # explicit definition
            ordered = self.__get_draw_strategy(ctx.draw_strategy())
            stv_list = self.visit(ctx.stvs())
            return ExplicitTargetDifficulty(stv_list, self.__default_reward_factory, ordered)

    def visitStv_pools(self, ctx: QrogueDungeonParser.Stv_poolsContext) -> None:
        for stv_pool in ctx.stv_pool():
            pool_id, target_difficulty = self.visit(stv_pool)
            self.__stv_pools[pool_id] = target_difficulty
        self.__default_target_difficulty = self.visit(ctx.default_stv_pool())
        self.__default_enemy_factory = EnemyFactory(self.__cbp.start_fight, self.__default_target_difficulty, 1)

    ##### Hallway area #####

    def visitH_attributes(self, ctx: QrogueDungeonParser.H_attributesContext) -> tiles.Door:
        direction = Direction.North  # todo adapt
        one_way_state = tiles.DoorOneWayState.NoOneWay
        if ctx.DIRECTION():
            dir_str = ctx.DIRECTION().getText()
            if dir_str == "North":
                direction = Direction.North
            elif dir_str == "East":
                direction = Direction.East
            elif dir_str == "South":
                direction = Direction.South
            elif dir_str == "West":
                direction = Direction.West
            if ctx.PERMANENT_LITERAL():
                one_way_state = tiles.DoorOneWayState.Permanent
            else:
                one_way_state = tiles.DoorOneWayState.Temporary

        ref_index = 0
        event_id = None
        if ctx.OPEN_LITERAL():
            open_state = tiles.DoorOpenState.Open
        elif ctx.CLOSED_LITERAL():
            open_state = tiles.DoorOpenState.Closed
        elif ctx.LOCKED_LITERAL():
            open_state = tiles.DoorOpenState.KeyLocked
        elif ctx.EVENT_LITERAL():
            if ctx.REFERENCE(ref_index):
                event_id = self.__normalize_reference(ctx.REFERENCE(ref_index).getText())
                open_state = tiles.DoorOpenState.EventLocked
                ref_index += 1
            else:
                open_state = tiles.DoorOpenState.Closed
                self.warning("Event lock specified without an event id! Ignoring the lock and placing a closed door "
                             "instead.")
        else:
            open_state = tiles.DoorOpenState.Closed
            self.warning("Invalid hallway attribute: it is neither locked nor opened nor closed!")

        # door = tiles.EntangledDoor(direction) # todo implement
        def door_check():
            return self.__check_achievement(event_id)
        door = tiles.Door(direction, open_state, one_way_state, door_check)

        if ctx.TUTORIAL_LITERAL():
            message = self.__load_message(ctx.REFERENCE(ref_index).getText())
            door.set_explanation(message)
            ref_index += 1
        if ctx.TRIGGER_LITERAL():
            event_to_trigger = self.__normalize_reference(ctx.REFERENCE(ref_index).getText())
            door.set_event(event_to_trigger)
        return door

    def visitHallway(self, ctx: QrogueDungeonParser.HallwayContext) -> Tuple[str, rooms.Hallway]:
        hw_id = ctx.HALLWAY_ID().getText()
        door = self.visit(ctx.h_attributes())
        return hw_id, door

    def visitHallways(self, ctx: QrogueDungeonParser.HallwaysContext) -> None:
        for hallway_ctx in ctx.hallway():
            hw_id, door = self.visit(hallway_ctx)
            self.__hallways_by_id[hw_id] = door
        # todo here we could entangle the doors if needed

    ##### Room area #####

    def visitShop_descriptor(self, ctx: QrogueDungeonParser.Shop_descriptorContext) -> tiles.ShopKeeper:
        if ctx.REFERENCE():
            pool_id = ctx.REFERENCE().getText()
            item_pool = self.__load_reward_pool(pool_id)
        else:
            item_pool = self.visit(ctx.collectibles())

        num_of_items = self.visit(ctx.integer())
        shop_factory = CollectibleFactory(item_pool)
        items = shop_factory.produce_multiple(self.__rm, num_of_items)
        return tiles.ShopKeeper(self.__cbp.visit_shop, [ShopItem(item) for item in items])

    def visitRiddle_descriptor(self, ctx: QrogueDungeonParser.Riddle_descriptorContext) -> tiles.Riddler:
        ref_index = 0
        if ctx.stv():
            stv = self.visit(ctx.stv())
        else:
            pool_id = ctx.REFERENCE(ref_index).getText()
            ref_index = 1
            difficulty = self.__load_stv_pool(pool_id)
            stv = difficulty.create_statevector(self.__robot, self.__rm)

        if ctx.collectible():
            reward = self.visit(ctx.collectible())
        else:
            pool_id = ctx.REFERENCE(ref_index).getText()
            reward_pool = self.__load_reward_pool(pool_id)
            reward = self.__rm.get_element(reward_pool)     # todo check if we should wrap it in a factory?

        riddle = Riddle(stv, reward)
        return tiles.Riddler(self.__cbp.open_riddle, riddle)

    def visitEnergy_descriptor(self, ctx: QrogueDungeonParser.Energy_descriptorContext) -> tiles.Tile:
        amount = self.visit(ctx.integer())
        return tiles.Energy(amount)

    def visitTrigger_descriptor(self, ctx: QrogueDungeonParser.Trigger_descriptorContext) -> tiles.Trigger:
        reference = ctx.REFERENCE().getText()
        return self.__load_trigger(reference)

    def visitMessage_descriptor(self, ctx: QrogueDungeonParser.Message_descriptorContext) -> tiles.Message:
        if ctx.integer():
            times = self.visit(ctx.integer())
        else:
            times = -1      # = always show
        reference = ctx.REFERENCE().getText()
        message = self.__load_message(reference)
        return tiles.Message(message, times)

    def visitCollectible_descriptor(self, ctx: QrogueDungeonParser.Collectible_descriptorContext) -> tiles.Collectible:
        pool_id = ctx.REFERENCE().getText()
        reward_factory = self.__load_reward_pool(pool_id)

        times = 1
        if ctx.integer():
            times = self.visit(ctx.integer())
        if times > 1:
            collectible = MultiCollectible(reward_factory.produce_multiple(self.__rm, times))
        else:
            collectible = reward_factory.produce(self.__rm)
        return tiles.Collectible(collectible)

    def visitEnemy_descriptor(self, ctx: QrogueDungeonParser.Enemy_descriptorContext) -> tiles.Enemy:
        enemy = None
        if self.__cur_room_id:
            room_id = self.__cur_room_id
        else:
            raise NotImplementedError()     # todo better check

        def get_entangled_tiles(id: int) -> List[tiles.Enemy]:
            if room_id in self.__enemy_groups_by_room:
                room_dic = self.__enemy_groups_by_room[room_id]
                return room_dic[id]
            else:
                return [enemy]

        enemy_id = int(ctx.DIGIT().getText())
        pool_id = ctx.REFERENCE(0).getText()
        if pool_id in self.__stv_pools:
            difficulty = self.__stv_pools[pool_id]
            enemy_factory = EnemyFactory(self.__cbp.start_fight, difficulty, 1)
            if ctx.REFERENCE(1):
                pool_id = ctx.REFERENCE(1).getText()
                if pool_id in self.__reward_pools:
                    reward_factory = self.__reward_pools[pool_id]
                    enemy_factory.set_custom_reward_factory(reward_factory)
        else:
            self.warning("Imports not yet supported! Choosing from default_stv_pool")
            enemy_factory = self.__default_enemy_factory

        enemy = tiles.Enemy(enemy_factory, get_entangled_tiles, enemy_id)
        if room_id not in self.__enemy_groups_by_room:
            self.__enemy_groups_by_room[room_id] = {}
        if enemy_id not in self.__enemy_groups_by_room[room_id]:
            self.__enemy_groups_by_room[room_id][enemy_id] = []
        self.__enemy_groups_by_room[room_id][enemy_id].append(enemy)
        return enemy

    def visitTile_descriptor(self, ctx: QrogueDungeonParser.Tile_descriptorContext) -> tiles.Tile:
        if ctx.trigger_descriptor():
            tile = self.visit(ctx.trigger_descriptor())
        elif ctx.enemy_descriptor():
            tile = self.visit(ctx.enemy_descriptor())
        elif ctx.collectible_descriptor():
            tile = self.visit(ctx.collectible_descriptor())
        elif ctx.energy_descriptor():
            tile = self.visit(ctx.energy_descriptor())
        elif ctx.riddle_descriptor():
            tile = self.visit(ctx.riddle_descriptor())
        elif ctx.shop_descriptor():
            tile = self.visit(ctx.shop_descriptor())
        elif ctx.message_descriptor():
            tile = self.visit(ctx.message_descriptor())
        else:
            self.warning("Invalid tile_descriptor! It is neither enemy, collectible, trigger or energy. "
                         "Returning tiles.Invalid() as consequence.")
            return tiles.Invalid()

        if isinstance(tile, tiles.WalkTriggerTile):
            ref_index = 0
            if ctx.TUTORIAL_LITERAL():
                ref = ctx.REFERENCE(ref_index).getText()
                msg = self.__load_message(ref)
                tile.set_explanation(msg)
                ref_index += 1
            if ctx.TRIGGER_LITERAL():
                ref = ctx.REFERENCE(ref_index).getText()
                event_id = self.__normalize_reference(ref)
                tile.set_event(event_id)
        return tile

    def visitTile(self, ctx: QrogueDungeonParser.TileContext) -> str:
        return ctx.getText()

    def visitR_row(self, ctx: QrogueDungeonParser.R_rowContext) -> List[str]:
        row = []
        for tile in ctx.tile():
            row.append(self.visit(tile))
        return row

    def visitR_type(self, ctx: QrogueDungeonParser.R_typeContext) -> rooms.AreaType:
        if ctx.SPAWN_LITERAL():
            return rooms.AreaType.SpawnRoom
        elif ctx.BOSS_LITERAL():
            return rooms.AreaType.BossRoom
        elif ctx.WILD_LITERAL():
            return rooms.AreaType.WildRoom
        elif ctx.SHOP_LITERAL():
            return rooms.AreaType.ShopRoom
        elif ctx.RIDDLE_LITERAL():
            return rooms.AreaType.RiddleRoom
        elif ctx.GATE_ROOM_LITERAL():
            return rooms.AreaType.GateRoom
        elif ctx.TREASURE_LITERAL():
            return rooms.AreaType.TreasureRoom
        else:
            self.warning(f"Invalid r_type: {ctx.getText()}")
            return rooms.AreaType.Invalid

    def visitR_visibility(self, ctx: QrogueDungeonParser.R_visibilityContext) -> Tuple[bool, bool]:
        visible = False
        foggy = False
        if ctx.VISIBLE_LITERAL():
            visible = True
        elif ctx.FOGGY_LITERAL():
            foggy = True
        return visible, foggy

    def visitR_attributes(self, ctx: QrogueDungeonParser.R_attributesContext) \
            -> Tuple[Tuple[bool, bool], rooms.AreaType]:
        visibility = self.visit(ctx.r_visibility())
        rtype = self.visit(ctx.r_type())
        return visibility, rtype

    def visitRoom(self, ctx: QrogueDungeonParser.RoomContext) -> Tuple[str, rooms.CustomRoom]:
        self.__cur_room_id = ctx.ROOM_ID().getText()
        visibility, room_type = self.visit(ctx.r_attributes())

        # place the tiles correctly in the room
        rows = []
        for row in ctx.r_row():
            rows.append(self.visit(row))
        tile_dic = {}
        for descriptor in ctx.tile_descriptor():
            tile = self.visit(descriptor)
            if tile.code is tiles.TileCode.Enemy:
                tile_str = str(tile.eid)
            else:
                tile_str = QrogueLevelGenerator.__tile_code_to_str(tile.code)
            if tile_str in tile_dic:
                tile_dic[tile_str].append(tile)
            else:
                tile_dic[tile_str] = [tile]

        # this local method is needed here for not explicitly defined enemies
        # but since we are already in a room we don't have to reference to global data
        enemy_dic = {}

        def get_entangled_tiles(eid: int) -> List[tiles.Enemy]:
            if eid in enemy_dic:
                return enemy_dic[eid]
            else:
                return []

        descriptor_indices = {}
        tile_matrix = []
        for row in rows:
            matrix_row = []
            for tile_str in row:
                if tile_str in tile_dic:
                    if tile_str not in descriptor_indices:
                        descriptor_indices[tile_str] = 0
                    index = descriptor_indices[tile_str]

                    tile = tile_dic[tile_str][index]
                    if index + 1 < len(tile_dic[tile_str]):     # todo maybe default tiles are better than using the last definition?
                        descriptor_indices[tile_str] = index + 1
                else:
                    tile = self.__get_default_tile(tile_str, enemy_dic, get_entangled_tiles)
                matrix_row.append(tile)
            # extended to the needed width with floors
            matrix_row += [tiles.Floor()] * (rooms.Room.INNER_WIDTH - len(matrix_row))
            tile_matrix.append(matrix_row)

        while len(tile_matrix) < rooms.Room.INNER_HEIGHT:
            tile_matrix.append([tiles.Floor()] * rooms.Room.INNER_WIDTH)

        if room_type is rooms.AreaType.SpawnRoom:
            tile_matrix[int(rooms.Room.INNER_HEIGHT / 2)][int(rooms.Room.INNER_WIDTH / 2)] = \
                tiles.Teleport(self.__teleport_callback, MapConfig.back_map_string(), None)

        # hallways will be added later
        room = rooms.CustomRoom(room_type, tile_matrix)
        visible, foggy = visibility
        if visible:
            room.make_visible()
        elif foggy:
            room.in_sight()
        return self.__cur_room_id, room

    def visitRooms(self, ctx: QrogueDungeonParser.RoomsContext):
        for room_ctx in ctx.room():
            room_id, room = self.visit(room_ctx)
            self.__rooms[room_id] = room

    ##### Layout area #####

    def visitLayout(self, ctx: QrogueDungeonParser.LayoutContext) -> List[List[rooms.Room]]:
        # first setup all hallway connections
        for y, hw_row in enumerate(ctx.l_hallway_row()):
            self.__visitL_hallway_row(hw_row, y)

        room_matrix = []
        for y in range(MapConfig.max_height()):
            row_ctx = ctx.l_room_row(y)
            if row_ctx:
                room_matrix.append(self.__visitL_room_row(row_ctx, y))
            else:
                break
        if ctx.l_room_row(MapConfig.max_height()):
            self.warning(f"Too much room rows specified. Only maps of size ({MapConfig.max_width()}, "
                         f"{MapConfig.max_height()}) supported. Ignoring over-specified rows.")

        return room_matrix

    def __hallway_handling(self, ctx_children: List[TerminalNodeImpl], y: int, direction: Direction):
        x = 0
        for child in ctx_children:
            if parser_util.check_for_overspecified_columns(x, child.symbol.type,
                                                           QrogueDungeonParser.VERTICAL_SEPARATOR):
                self.warning(
                    f"Too much room columns specified. Only maps of size ({MapConfig.max_width()}, "
                    f"{MapConfig.max_height()}) supported. Ignoring over-specified columns.")
                break
            if child.symbol.type == QrogueDungeonParser.HALLWAY_ID:
                hw_id = child.symbol.text
                origin = Coordinate(x, y)
                self._add_hallway(origin, origin + direction, self.__load_hallway(hw_id))
                x += 1
            elif child.symbol.type == QrogueDungeonParser.EMPTY_HALLWAY:
                x += 1

    def __visitL_hallway_row(self, ctx: QrogueDungeonParser.L_hallway_rowContext, y: int) -> None:
        self.__hallway_handling(ctx.children, y, Direction.South)    # connect downwards to the next room row

    def __visitL_room_row(self, ctx: QrogueDungeonParser.L_room_rowContext, y: int) -> List[rooms.Room]:
        self.__hallway_handling(ctx.children, y, Direction.East)     # connect to the right to the next room

        row = []
        x = 0
        for child in ctx.children:
            if parser_util.check_for_overspecified_columns(x, child.symbol.type,
                                                           QrogueDungeonParser.VERTICAL_SEPARATOR):
                self.warning(f"Too much room columns specified. Only maps of size ({MapConfig.max_width()}, "
                             f"{MapConfig.max_height()}) supported. Ignoring over-specified columns.")
                break

            if child.symbol.type == QrogueDungeonParser.ROOM_ID:
                room_id = child.symbol.text     # todo make it illegal to have the same room_id twice?
                row.append(self.__load_room(room_id, x, y))
                x += 1
            elif child.symbol.type == QrogueDungeonParser.EMPTY_ROOM:
                row.append(None)
                x += 1
        return row

    ##### Robot area #####

    def visitRobot(self, ctx: QrogueDungeonParser.RobotContext) -> None:
        num_of_qubits = int(ctx.DIGIT().getText())
        gates = []
        for ref in ctx.REFERENCE():
            if parser_util.normalize_reference(ref.getText()) == self.__ROBOT_NO_GATES:
                continue
            gates.append(self.__load_gate(ref.getText()))  # todo what about pickups?

        self.__robot = TestBot(CallbackPack.instance().game_over, num_of_qubits, gates)

    ##### Start area #####

    def visitStart(self, ctx: QrogueDungeonParser.StartContext) -> Tuple[str, List[List[rooms.Room]]]:
        if ctx.NAME():
            name = ctx.TEXT().getText()[1:-1]
        else:
            name = None

        # prepare the robot
        self.visit(ctx.robot())

        # prepare messages
        self.visit(ctx.messages())

        # prepare reward pools first because they are standalone
        self.visit(ctx.reward_pools())
        # prepare state vector pools next because they can only reference reward pools
        self.visit(ctx.stv_pools())

        # prepare hallways next because they are standalone but thematically more connected to rooms
        self.visit(ctx.hallways())
        # now prepare the rooms because they can reference everything above
        self.visit(ctx.rooms())

        # for the last step we retrieve the room matrix from layout
        return name, self.visit(ctx.layout())



# exporting
import qrogue.game.logic
import qrogue.game.world

from .target_factory import TargetDifficulty, ExplicitTargetDifficulty, \
    BossFactory, EnemyFactory, RiddleFactory

# importing
# +util



from enum import Enum

from qrogue.util.config import ColorConfig as CC


class _HL:
    # objects
    boss = CC.highlight_object("Boss")
    circuit = CC.highlight_object("Circuit")
    coins = CC.highlight_object("Coins")
    collec = CC.highlight_object("Collectibles")
    door = CC.highlight_object("Door")
    enemies = CC.highlight_object("Enemies")
    energy = CC.highlight_object("Energy")
    gates = CC.highlight_object("Gates")
    key = CC.highlight_object("Key")
    keys = CC.highlight_object("Keys")

    quantum_circuit = CC.highlight_object("Quantum Circuit")
    quantum_gates = CC.highlight_object("Quantum Gates")
    quantum_state = CC.highlight_object("Quantum State")
    qubits = CC.highlight_object("Qubits")
    qubit_s = CC.highlight_object("Qubit(s)")
    riddles = CC.highlight_object("Riddles")
    robot = CC.highlight_object("Robot")
    shop = CC.highlight_object("Shop")
    special_rooms = CC.highlight_object("Special Rooms")
    state = CC.highlight_object("State")
    state_vectors = CC.highlight_object("StateVectors")
    target_state = CC.highlight_object("Target State")

    # words
    abort = CC.highlight_word("abort")
    action = CC.highlight_word("action")
    add_remove = CC.highlight_word("Add/remove")
    attempt = CC.highlight_word("Attempt")
    bell = CC.highlight_word("Bell, the Master of Entanglement")
    both0 = CC.highlight_word("both 0")
    both1 = CC.highlight_word("both 1")
    buy = CC.highlight_word("buy")

    cannot_flee = CC.highlight_word("cannot flee")
    chance_based = CC.highlight_word("chance based")
    commit = CC.highlight_word("Commit")
    continue_ = CC.highlight_word("Continue")
    details = CC.highlight_word("details")
    difference = CC.highlight_word("Difference")
    display = CC.highlight_word("display")
    display_details = CC.highlight_word("displays details")
    easier = CC.highlight_word("easier")
    enemy = CC.highlight_object("Enemy")
    entanglement = CC.highlight_word("Entanglement")
    exit_ = CC.highlight_word("Exit")
    fights = CC.highlight_word("Fights")
    fight = CC.highlight_word("Fight")
    five_parenthesis = CC.highlight_word("5)")
    flee = CC.highlight_word("Flee")
    four_parenthesis = CC.highlight_word("4)")

    gate = CC.highlight_object("Gate")
    give_up = CC.highlight_word("\"Give up\"")
    help_ = CC.highlight_word("Manual")
    hp = CC.highlight_word("don't lose HP")
    items = CC.highlight_word("Items")
    leave = CC.highlight_word("\"-Leave-\"")
    level = CC.highlight_word("Level")
    list_ = CC.highlight_word("list")
    locked = CC.highlight_word("locked")
    map_ = CC.highlight_word("map")
    navigate = CC.highlight_word("navigate")
    not_zero = CC.highlight_word("not zero")
    one_hp = CC.highlight_word("1 HP")
    one_parenthesis = CC.highlight_word("1)")
    options = CC.highlight_word("Options")

    quantum_computing = CC.highlight_word("Quantum Computing")
    quantum_algorithm = CC.highlight_word("Quantum Algorithm")
    reenter = CC.highlight_word("re-enter")
    removed = CC.highlight_word("removed")
    reward = CC.highlight_word("Reward")
    serious = CC.highlight_word("serious")
    two_parenthesis = CC.highlight_word("2)")
    three_parenthesis = CC.highlight_word("3)")
    use = CC.highlight_word("use")
    vanishes = CC.highlight_word("vanishes")
    win = CC.highlight_word("win")
    zeros = CC.highlight_word("zeros")

    # keys
    arrow_keys = CC.highlight_key("Arrow Keys")
    backspace = CC.highlight_key("Backspace")
    ctrl_q = CC.highlight_key("CTRL+Q")
    p = CC.highlight_key("P")
    q = CC.highlight_key("Q")
    enter = CC.highlight_key("Enter")
    escape = CC.highlight_key("ESC")
    shift_a = CC.highlight_key("Shift+A")
    shortcuts = CC.highlight_key("0") + ", " + CC.highlight_key("1") + ", ... , " + CC.highlight_key("9")
    space = CC.highlight_key("Space")
    tab = CC.highlight_key("TAB")
    wasd = CC.highlight_key("WASD")

    # tiles
    DOOR_TILE = CC.highlight_tile("-")
    TUTORIAL_TILE = CC.highlight_tile(".")


class HelpTextType(Enum):
    Controls = 0
    Fight = 1
    Riddle = 2
    Shop = 3
    BossFight = 4
    Game = 5
    Pause = 6
    Options = 7
    Welcome = 8
    FirstLevelIntroduction = 9


class HelpText:
    __DIC = {
        HelpTextType.Welcome:
            f"Qrogue is a game about {_HL.quantum_computing}. You will explore "
            f"Dungeons with the help of {_HL.quantum_gates} you can use for your "
            f"{_HL.quantum_circuit}. But you are not the only one in the Dungeons! "
            f"There are {_HL.enemies} challenging you to reach a certain "
            f"{_HL.quantum_state}. Your goal is to expand your library of "
            f"{_HL.quantum_gates} which are hidden in "
            f"{_HL.special_rooms} in the Dungeon or guarded "
            f"by a {_HL.boss} - a special Enemy that wants to see a "
            f"{_HL.quantum_algorithm} from you...\n"
            f"Now let's start! Try to move around with the {_HL.arrow_keys} and "
            f"go to the {_HL.door} ({_HL.DOOR_TILE}) at the bottom!"
            f"\nThe fields with a {_HL.TUTORIAL_TILE} will show you the next steps. "
            f"Now close this dialog and start playing by pressing {_HL.space}.",

        HelpTextType.FirstLevelIntroduction:
            f"Alright, let's have a look at our {_HL.display}.\n"
            f"The HUD is at the very top and shows you the current {_HL.level}, the [{_HL.energy} of the {_HL.robot} and the number "
            f"of {_HL.coins} and {_HL.keys} it is carrying. Always keep an eye on your current energy as 0 energy means we can no "
            f"longer control the {_HL.robot} and have to {_HL.abort} the {_HL.level}.\n"
            f"The rest of the screen shows a {_HL.map_} of the area our Robot's currently in. Try to move the Robot to the "
            f"top left corner of the current room.",

        HelpTextType.Controls:
            "Don't forget to scroll down to see all controls!\n"
            f"Move                  -   {_HL.arrow_keys}, {_HL.wasd}\n"
            f"Navigate Menus        -   {_HL.arrow_keys}, {_HL.wasd}\n"
            f"Confirm               -   {_HL.enter}, {_HL.space}\n"
            f"Cancel/Back           -   {_HL.backspace}, {_HL.shift_a}\n"
            f"Scroll in Popup       -   {_HL.arrow_keys}, {_HL.wasd}\n"
            f"Close Popup           -   {_HL.enter}, {_HL.space}, {_HL.escape}\n"
            f"Pause                 -   {_HL.p}, {_HL.tab}\n"
            f"Selection shortcuts   -   {_HL.shortcuts}\n"
            #"\n"
            f"Should you ever get stuck you can force-quit the game by pressing {_HL.ctrl_q} and then {_HL.q}. This "
            "will still save the logs so it is the preferred option over simply closing the window!",

        HelpTextType.Fight:
            f"{_HL.one_parenthesis} In the middle of the screen you see 3 {_HL.state_vectors}. The left one (Current "
            f"State) corresponds to the output of your {_HL.circuit} (more on that later) while the right one "
            f"(Target State) is constant and depending on the "
            f"{_HL.enemy} you fight. You {_HL.win} the Fight by setting your Current State to the Target State by "
            f"adapting the mentioned Circuit. In between those two you can see their {_HL.difference}. If "
            f"it shows only {_HL.zeros} you reached your target and won the Fight!\n"
            f"{_HL.two_parenthesis} Underneath the StateVectors is your {_HL.circuit}. Currently you have 2 "
            f"{_HL.qubits} (q0, q1) and 0 out of 3 {_HL.gates} applied to them. The before mentioned Current "
            f"{_HL.state} reflects the output (out) of your {_HL.circuit}.\n"
            f"{_HL.three_parenthesis} On the bottom left you can choose the {_HL.action} you want to take: \n"
            f"{_HL.add_remove} - Change your {_HL.circuit} with the {_HL.gates} available to you (selection to the right). "
            f"After selecting a {_HL.gate} you are asked on which {_HL.qubit_s} you want to place it. If you select an "
            f"already used one it will be {_HL.removed} from your Circuit instead.\n"
            f"{_HL.commit} - Commit your changes and update your {_HL.circuit}. If Difference is {_HL.not_zero} you "
            f"lose {_HL.one_hp}.\n"
            f"{_HL.items} - Use one of your Items to make the Fight {_HL.easier} (you don't have any Items yet!)\n"
            f"{_HL.flee} - Try to flee from the Fight. This is {_HL.chance_based} and you lose {_HL.one_hp} if you "
            f"fail to flee (Note: for Tutorial purposes you {_HL.cannot_flee} in this Room!)\n"
            f"{_HL.four_parenthesis} The bottom right {_HL.display_details} based on the action you chose on the left "
            f"side. E.g. you can select the {_HL.gate} you want to use in your {_HL.circuit}.\n"
            f"{_HL.five_parenthesis} Use your {_HL.arrow_keys} to {_HL.navigate} between your available options at the "
            f"bottom and press {_HL.space} to {_HL.use} the selected one. Again, your goal now is to reach the "
            f"{_HL.target_state} of the {_HL.enemy}. If you succeed, you will get a {_HL.reward}!",

        HelpTextType.Riddle:
            f"{_HL.riddles} are very similar to {_HL.fights}. You have a {_HL.target_state} you need to reach (Difference is zero) "
            f"by adapting your {_HL.circuit}. The main difference is that you {_HL.hp} if you fail but instead 1 "
            f"{_HL.attempt} for solving the Riddle. When you have no more Attempts left the Riddle {_HL.vanishes} "
            "together with its reward - which is usually much better than the ones from Fights. Also fleeing (or in "
            f"this case {_HL.give_up}) will always work but obviously cost you your current {_HL.attempt} which is "
            f"why you are notified if this would lead to 0 Attempts left.",

        HelpTextType.Shop:
            f"In the {_HL.shop} you can use {_HL.coins} you got (e.g. from Fights) to buy various {_HL.collec}. On the "
            f"left side is a {_HL.list_} of everything you can {_HL.buy}. Navigate as usual with your {_HL.arrow_keys} "
            f"and select something with {_HL.space} to see more {_HL.details} on the right side. There you can also "
            f"buy it.\n"
            f"{_HL.leave} obviously makes you leave the {_HL.shop}. You can always {_HL.reenter} it later if you want!",

        HelpTextType.BossFight: # todo make more generic
            f"Now it's getting {_HL.serious}! You are fighting against {_HL.bell}. For the {_HL.state} you need to reach to "
            f"defeat Bell your two {_HL.qubits} will always have to be the same: either {_HL.both0} or {_HL.both1}.\n"
            f"This is called {_HL.entanglement}.\n\n"
            "Good luck!",

        HelpTextType.Game:
            f"Qrogue is a game about {_HL.quantum_computing}. You will explore "
            f"Dungeons with the help of {_HL.quantum_gates} you can use for your "
            f"{_HL.quantum_circuit}. But you are not the only one in the Dungeons! "
            f"There are {_HL.enemies} challenging you to reach a certain "
            f"{_HL.quantum_state}. Your goal is to expand your library of "
            f"{_HL.quantum_gates} which are hidden in "
            f"{_HL.special_rooms} in the Dungeon or guarded "
            f"by a {_HL.boss} - a special Enemy that wants to see a "
            f"{_HL.quantum_algorithm} from you.\n",

        HelpTextType.Pause:
            "In the Pause Menu you can do several things:\n"
            f"{_HL.continue_} - Leave the Pause Menu and continue where you stopped.\n"
            f"{_HL.help_} - If you ever feel stuck and don't remember how certain stuff in the game works select "
            f"this menu and we will try to help you.\n"
            f"{_HL.options} - Configure some Options of the game, like font size or coloring.\n"
            f"{_HL.exit_} - Exit the current Level or Expedition and return to the Spaceship.\n",

        HelpTextType.Options:
            "",

    }

    @staticmethod
    def get(type: HelpTextType) -> str:
        return HelpText.__DIC[type]

    @staticmethod
    def load(type: str) -> str:
        for key in HelpText.__DIC.keys():
            if key.name.lower() == type.lower():
                return HelpText.__DIC[key]
        return None


class TutorialTextType(Enum):
    LockedDoorNoKey = 0
    LockedDoorKey = 1


class TutorialText:
    __DIC = {
        TutorialTextType.LockedDoorNoKey:
            f"Hmm, the {_HL.door} is {_HL.locked}. Let's see if we can find a {_HL.key} somewhere.",
        TutorialTextType.LockedDoorKey:
            f"Oh the {_HL.door} is {_HL.locked}. Luckily we've already found a {_HL.key} to open it!",
    }

    @staticmethod
    def get(type: TutorialTextType) -> str:
        return TutorialText.__DIC[type]

    @staticmethod
    def load(type: str) -> str:
        for key in TutorialText.__DIC.keys():
            if key.name.lower() == type.lower():
                return TutorialText.__DIC[key]
        return None



from typing import List, Callable

from qrogue.game.logic.actors import StateVector
from qrogue.game.logic.actors.controllables import Robot
from qrogue.game.logic.actors.puzzles import DummyEnemy, Enemy, Target, Riddle, Boss
from qrogue.game.logic.collectibles import Collectible, CollectibleFactory, Instruction, CXGate, SwapGate, HGate, \
    XGate, Coin, Key
from qrogue.game.world.navigation import Direction
from qrogue.util import Logger, MyRandom, RandomManager


class TargetDifficulty:
    """
    A class that handles all parameters that define the difficulty of a fight.
    """

    def __init__(self, num_of_instructions: int, rewards):
        """

        :param num_of_instructions: number of Instructions used to create a target StateVector
        :param rewards: either a list of Collectibles or a CollectibleFactory for creating a reward when reaching
        a Target
        """
        self.__num_of_instructions = num_of_instructions
        if isinstance(rewards, list):
            self.__reward_factory = CollectibleFactory(rewards)
        elif isinstance(rewards, CollectibleFactory):
            self.__reward_factory = rewards
        else:
            Logger.instance().throw(ValueError(
                "rewards must be either a list of Collectibles or a CollectibleFactory"))

    def produce_reward(self, rm: MyRandom):
        return self.__reward_factory.produce(rm)

    def create_statevector(self, robot: Robot, rm: MyRandom) -> StateVector:
        """
        Creates a random StateVector that is reachable for the given Robot.

        :param robot: provides the needed information regarding the number of qubits and usable Instructions for
        creating a StateVector
        :param rm: seeded randomness for choosing Instructions and the Qubit(s) to use them on
        :return: a StateVector reachable for the provided Robot
        """
        num_of_qubits = robot.num_of_qubits

        # choose random circuits on random qubits and cbits
        instruction_pool = robot.get_available_instructions()
        instructions = []
        num_of_instructions = min(self.__num_of_instructions, robot.circuit_space)
        for i in range(num_of_instructions):
            qubits = list(range(num_of_qubits))
            # remove = len(instruction_pool) - (self.__num_of_instructions - i) >= 0
            # if not remove:
            #     Logger.instance().throw(Exception(
            #         "this should always remove because else we would duplicate instructions"))
            instruction = rm.get_element(instruction_pool, remove=True)
            while instruction.use_qubit(rm.get_element(qubits, remove=True)):
                pass
            instructions.append(instruction)
        return StateVector.from_gates(instructions, num_of_qubits)


class ExplicitTargetDifficulty(TargetDifficulty):
    """
    A TargetDifficulty that doesn't create StateVectors based on a Robot's possibilities but by choosing from a pool
    of explicitely provided StateVectors
    """

    def __init__(self, stv_pool: List[StateVector], reward_factory: CollectibleFactory, ordered: bool = False):
        """

        :param stv_pool: list of StateVectors to choose from
        :param reward_factory: factory for creating a reward
        :param ordered: whether StateVectors should be chosen in order or randomly from the given stv_pool
        """
        super().__init__(-1, reward_factory)
        self.__pool = stv_pool
        self.__ordered = ordered
        self.__order_index = -1

    def create_statevector(self, robot: Robot, rm: MyRandom) -> StateVector:
        if self.__ordered or rm is None:
            self.__order_index += 1
            if self.__order_index >= len(self.__pool):
                self.__order_index = 0
            return self.__pool[self.__order_index]
        else:
            return rm.get_element(self.__pool)


class RiddleDifficulty(TargetDifficulty):
    def __init__(self, num_of_instructions: int, reward_pool: "list of Collectibles", min_attempts: int = 1,
                 max_attempts: int = 10):
        super().__init__(num_of_instructions, reward_pool)
        self.__min_attempts = min_attempts
        self.__max_attempts = max_attempts

    def get_attempts(self, rm: MyRandom) -> int:
        return rm.get_int(self.__min_attempts, self.__max_attempts)


class DummyTargetDifficulty(TargetDifficulty):
    """
    Dummy implementation of a FightDifficulty class for testing.
    """

    def __init__(self):
        super(DummyTargetDifficulty, self).__init__(2, [Coin(1), Coin(3)])


class EnemyFactory:
    """
    This class produces enemies (actors) with a certain difficulty.
    It is used by enemy tiles to trigger a fight.
    """

    def __init__(self, start_fight_callback: Callable[[Robot, Enemy, Direction], None], difficulty: TargetDifficulty,
                 default_flee_chance: float = 0.5):
        """

        :param difficulty: difficulty of the enemy we produce
        """
        self.__start_fight = start_fight_callback
        self.__difficulty = difficulty
        self.__default_flee_chance = default_flee_chance
        self.__custom_reward_factory = None

    def set_custom_reward_factory(self, factory: CollectibleFactory):
        self.__custom_reward_factory = factory

    def produce(self, robot: Robot, rm: MyRandom, flee_chance: float = None) -> Enemy:
        """
        Creates an enemy based on the number of qubits the provided robot has.

        :param robot:
        :param rm:
        :param flee_chance: chance of the robot to flee from the fight
        :return: a freshly created enemy
        """
        stv = self.__difficulty.create_statevector(robot, rm)
        if self.__custom_reward_factory:
            reward = self.__custom_reward_factory.produce(rm)
        else:
            reward = self.__difficulty.produce_reward(rm)
        if flee_chance is None:
            flee_chance = self.__default_flee_chance
        return DummyEnemy(stv, reward, flee_chance)

    def start(self, robot: Robot, enemy: Enemy, direction: Direction):
        self.__start_fight(robot, enemy, direction)


class ExplicitEnemyFactory(EnemyFactory):
    def __init__(self, start_fight_callback: Callable[[Robot, Target, Direction], None], stv_pool: [StateVector],
                 reward_pool: [Collectible]):
        self.__stv_pool = stv_pool
        self.__reward_pool = reward_pool
        super().__init__(start_fight_callback, DummyTargetDifficulty())

    def produce(self, robot: Robot, rm: MyRandom, flee_chance: float) -> Enemy:
        stv = rm.get_element(self.__stv_pool)
        reward = rm.get_element(self.__reward_pool)
        return DummyEnemy(stv, reward, flee_chance)


class RiddleFactory:
    @staticmethod
    def default(robot: Robot) -> "RiddleFactory":
        reward_pool = [CXGate(), HGate, XGate, SwapGate, Coin(11), Key(5)]
        difficulty = RiddleDifficulty(num_of_instructions=4, reward_pool=reward_pool, min_attempts=4, max_attempts=9)
        return RiddleFactory(robot, difficulty)

    def __init__(self, robot: Robot, difficulty: RiddleDifficulty):
        self.__robot = robot
        self.__difficulty = difficulty

    def produce(self, rm: MyRandom) -> Riddle:
        stv = self.__difficulty.create_statevector(self.__robot, rm)
        reward = self.__difficulty.produce_reward(rm)
        attempts = self.__difficulty.get_attempts(rm)
        return Riddle(stv, reward, attempts)


class BossFactory:
    @staticmethod
    def default(robot: Robot) -> "BossFactory":
        pool = [CXGate(), SwapGate(), Coin(30)]
        return BossFactory(robot, pool)

    def __init__(self, robot: Robot, reward_pool: List[Collectible]):
        self.__robot = robot
        self.__reward_pool = reward_pool
        self.__rm = RandomManager.create_new()

    def produce(self, include_gates: List[Instruction]) -> Boss:
        used_gates = []
        qubit_count = [0] * self.__robot.num_of_qubits
        qubits = list(range(self.__robot.num_of_qubits))

        for g in include_gates:
            gate = g.copy()
            self.__prepare_gate(gate, qubit_count, qubits)
            used_gates.append(gate)

        usable_gates = self.__robot.get_available_instructions()
        while len(usable_gates) > 0:
            gate = self.__rm.get_element(usable_gates, remove=True)
            self.__prepare_gate(gate, qubit_count, qubits)
            used_gates.append(gate)

        reward = self.__rm.get_element(self.__reward_pool)
        return Boss(StateVector.from_gates(used_gates, self.__robot.num_of_qubits), reward)

    def __prepare_gate(self, gate: Instruction, qubit_count, qubits):
        gate_qubits = qubits.copy()
        while True:
            qubit = self.__rm.get_element(gate_qubits, remove=True)
            qubit_count[qubit] += 1
            if qubit_count[qubit] >= self.__robot.circuit_space:
                qubits.remove(qubit)
            if not gate.use_qubit(qubit):
                return
